import pandas as pd
import numpy as np
from packaging import version

from ... import units as u
from ... import utils
from ... import settings
from ... base.types import FLOAT, INT


def convert_data(data, unit, to_unit):
    """
    Converts data to desired unit or leaves it dimensionless.

    :param data: numpy.array; data array to be converted
    :param unit: astropy.unit; `data` unit
    :param to_unit: astropy.unit; unit into which the `data` will be transformed
    :return: numpy.array; converted array
    """
    return data if unit == u.dimensionless_unscaled else (data * unit).to(to_unit).value


def convert_flux(data, unit, zero_point=None):
    """
    If the input flux is in magnitudes, they are converted to normalized flux.

    :param data: numpy.array; flux array
    :param unit: astropy.unit.Unit; flux unit of `array` (dimensionless or magnitudes)
    :param zero_point: float; reference magnitude of the dataset, in case, the flux was provided in magnitudes
    :return: numpy.array; converted normalized flux
    """
    if unit in [u.mag, u.mmag] and zero_point is None:
        raise ValueError('You supplied your data in magnitudes. Please also specify '
                         'a zero point using keyword argument `reference_magnitude`.')

    if unit == u.mag:
        data = utils.magnitude_to_flux(data, zero_point)
    elif unit == u.mmag:
        data = utils.magnitude_to_flux(data/1000.0, zero_point)

    return data


def convert_flux_error(error, unit, zero_point=None):
    """
    If data and its errors are in magnitudes, they are converted to normalized flux.

    :param error: numpy.array; flux error array
    :param unit: astropy.unit.Unit; flux unit of `error` (dimensionless or magnitudes)
    :param zero_point: float; float; reference magnitude of the dataset, in case, the flux was provided in magnitudes
    :return: numpy.array; converted error in normalized flux
    """
    if unit in [u.mag, u.mmag] and zero_point is None:
        raise ValueError('You supplied your data in magnitudes. Please also specify '
                         'a zero point using keyword argument `reference_magnitude`.')

    if unit == u.mag:
        error = utils.magnitude_error_to_flux_error(error)
    elif unit == u.mmag:
        error = utils.magnitude_error_to_flux_error(error/1000.0)
    return error


def convert_unit(unit, to_unit):
    """
    Replacing unit by desired unit or leave it dimensionless.

    :param unit: astropy.unit; old unit
    :param to_unit: astropy.unit; new unit
    :return: astropy.unit; new unit
    """
    return unit if unit == u.dimensionless_unscaled else to_unit


def read_data_file(filename, data_columns, delimiter=settings.DELIM_WHITESPACE):
    """
    Function loads observation datafile. Rows with column names and comments should start with `#`.
    It deals with missing data by omitting given line.

    :param delimiter: str; regex to define columns separtor
    :param filename: str;
    :param data_columns: Tuple; (time column, observable column, observable error column)
    :return: numpy.array; (N x 3) matrix containing loaded data in columns
    """
    reader_kwargs = dict(on_bad_lines='skip')
    if version.parse(pd.__version__) < version.parse('1.3.0'):
        reader_kwargs = dict(error_bad_lines=False)

    data = pd.read_csv(
        filename, header=None, comment='#', delimiter=delimiter,
        engine='python', **reader_kwargs
    )[list(data_columns)]
    data = data.apply(lambda s: pd.to_numeric(s, errors='coerce')).dropna()
    return data.to_numpy(dtype=FLOAT)


def central_moving_average(dt_set, n_bins=100, radius=2, cyclic_boundaries=True):
    """
    Function performs central moving averages in order to smooth observations in given
    dataset. The method divides the phase curve into `n_bins`. Afterward, for each bin, the average flux is
    calculated for points within `radius` number of bins. Use this function only on phased data.

    :param dt_set: numpy.array; phase curve to be smoothed
    :param n_bins: int; number of bins on the phase curve
    :param radius: int; amount of bins from which the average is calculated
    :param cyclic_boundaries: bool; last bin is direct neighbour of the first bin
    """
    bin_boundaries = np.linspace(dt_set.x_data.min(), dt_set.x_data.max(), num=n_bins + 1, endpoint=True)
    bin_centres = 0.5 * (bin_boundaries[:-1] + bin_boundaries[1:])
    bin_idxs = np.digitize(dt_set.x_data, bin_boundaries[1:], right=True)

    if cyclic_boundaries:
        bins = [(np.arange(start=ii-radius, stop=ii+radius+1, step=1.0, dtype=INT) % n_bins)
                for ii in range(n_bins)]
    else:
        bins = [np.arange(start=0, stop=ii+radius+1, step=1.0, dtype=INT) for ii in range(radius)]
        bins += [np.arange(start=ii-radius, stop=ii+radius+1, step=1.0, dtype=INT)
                 for ii in range(radius, n_bins-radius)]
        bins += [np.arange(start=ii-radius, stop=n_bins, step=1.0, dtype=INT)
                 for ii in range(n_bins-radius, n_bins)]

    bin_masks = [np.isin(bin_idxs, bins[ii]) for ii in range(n_bins)]
    bin_masks_count = np.sum(bin_masks, axis=1)

    non_empty_bins = bin_masks_count > 0
    iterator = np.arange(n_bins)[non_empty_bins]

    if dt_set.y_err is not None:
        bin_averages = np.array([np.average(dt_set.y_data[bin_masks[ii]],
                                            weights=1/dt_set.y_err[bin_masks[ii]]**2)
                                 for ii in iterator])
        bin_errors = np.array([np.mean(dt_set.y_err[bin_masks[ii]]) for ii in iterator])
    else:
        bin_averages = np.array([np.average(dt_set.y_data[bin_masks[ii]]) for ii in iterator])
        bin_errors = np.array([np.std(dt_set.y_data[bin_masks[ii]]) for ii in iterator(n_bins)])

    dt_set.x_data = bin_centres[non_empty_bins]
    dt_set.y_data = bin_averages
    dt_set.y_err = bin_errors



import numpy as np

from .. import utils as bsutils
from ... utils import is_empty
from ... logger import getLogger
from ... base.surface import gravity as bgravity
from ... import (
    umpy as up,
    const
)

logger = getLogger("binary_system.surface.gravity")


def calculate_potential_gradient(components_distance, component, points, synchronicity, mass_ratio):
    """
    Return outter gradients in each point of star surface or defined points.
    If points are not supplied, component instance points are used.

    :param component: str; define target component to compute critical potential; `primary` or `secondary`
    :param components_distance: float, in SMA distance
    :param points: numpy.array;
    :param synchronicity: float;
    :param mass_ratio: float;
    :return: numpy.array;
    """
    r3 = up.power(np.linalg.norm(points, axis=1), 3)
    r_hat3 = up.power(np.linalg.norm(points - np.array([components_distance, 0., 0]), axis=1), 3)

    f2 = up.power(synchronicity, 2)
    if component == 'primary':
        domega_dx = - points[:, 0] / r3 + \
                    mass_ratio * (components_distance - points[:, 0]) / r_hat3 + \
                    f2 * (mass_ratio + 1) * points[:, 0] - \
                    mass_ratio / up.power(components_distance, 2)
    elif component == 'secondary':
        domega_dx = - points[:, 0] / r3 + \
                    mass_ratio * (components_distance - points[:, 0]) / r_hat3 - \
                    f2 * (mass_ratio + 1) * (components_distance - points[:, 0]) + \
                    1 / up.power(components_distance, 2)
    else:
        raise ValueError(f'Invalid value `{component}` of argument `component`.\n Use `primary` or `secondary`.')
    domega_dy = - points[:, 1] * (1 / r3 + mass_ratio / r_hat3 - f2 * (mass_ratio + 1))
    domega_dz = - points[:, 2] * (1 / r3 + mass_ratio / r_hat3)
    return -np.column_stack((domega_dx, domega_dy, domega_dz))


def calculate_polar_potential_gradient_magnitude(components_distance, mass_ratio, polar_radius, component,
                                                 synchronicity):
    """
    Calculate magnitude of polar potential gradient.

    :param components_distance: float; in SMA distance
    :param polar_radius: float;
    :param mass_ratio: float;
    :param component: str;
    :param synchronicity: float;
    :return: float;
    """
    points = [0., 0., polar_radius] if component == 'primary' else [components_distance, 0., polar_radius]
    points = np.array(points)
    r3 = up.power(np.linalg.norm(points), 3)
    r_hat3 = up.power(np.linalg.norm(points - np.array([components_distance, 0., 0.])), 3)

    if component == 'primary':
        domega_dx = mass_ratio * components_distance / r_hat3 - mass_ratio / up.power(components_distance, 2)
    elif component == 'secondary':
        domega_dx = - points[0] / r3 + mass_ratio * (components_distance - points[0]) / r_hat3 \
                    - np.power(synchronicity, 2) * (mass_ratio + 1) * (1 - points[0])\
                    + 1. / up.power(components_distance, 2)
    else:
        raise ValueError(f'Invalid value `{component}` of argument `component`. \nUse `primary` or `secondary`.')
    domega_dz = - points[2] * (1. / r3 + mass_ratio / r_hat3)
    return up.power(up.power(domega_dx, 2) + up.power(domega_dz, 2), 0.5)


def calculate_polar_gravity_acceleration(star, components_distance, mass_ratio, component,
                                         semi_major_axis, synchronicity, logg=False):
    """
    Calculates polar gravity acceleration for component of binary system.
    Calculated from gradient of Roche potential::

        d_Omega/dr using transformation g = d_Psi/dr = (GM_component/semi_major_axis**2) * d_Omega/dr
        ( * 1/q in case of secondary component )

    :param star: elisa.base.container.StarContainer;
    :param components_distance: float; (in SMA units)
    :param mass_ratio: float;
    :param component: str;
    :param semi_major_axis: float;
    :param logg: bool; if True log g is returned, otherwise values are not in log10
    :param synchronicity: float;
    :return: numpy.array; surface gravity or log10 of surface gravity
    """
    pgm = calculate_polar_potential_gradient_magnitude(components_distance, mass_ratio,
                                                       star.polar_radius, component, synchronicity)
    gradient = const.G * star.mass * pgm / up.power(semi_major_axis, 2)
    gradient = gradient / mass_ratio if component == 'secondary' else gradient
    return up.log10(gradient) if logg else gradient


def build_surface_gravity(system, components_distance, component="all"):
    """
    Function calculates gravity potential gradient magnitude (surface gravity) for each face.
    Value assigned to face is calculated as a mean of surface gravity values calculated in corners of given face.

    :param system: elisa.binary_system.container.OrbitalPositionContainer;
    :param component: str; `primary` or `secondary`
    :param components_distance: float;
    :return: system: elisa.binary_system.container.OrbitalPositionContainer;
    """
    if is_empty(component):
        logger.debug("no component set to build surface gravity")
        return

    if is_empty(components_distance):
        raise ValueError('Component distance value was not supplied or is invalid.')

    components = bsutils.component_to_list(component)
    mass_ratio = system.mass_ratio

    for component in components:
        star = getattr(system, component)
        synchronicity = star.synchronicity

        pgm = calculate_polar_potential_gradient_magnitude(components_distance, mass_ratio,
                                                           star.polar_radius, component, star.synchronicity)
        setattr(star, "polar_potential_gradient_magnitude", pgm)

        logger.debug(f'computing potential gradient magnitudes distribution of {component} component')

        points, faces = bgravity.eval_args_for_magnitude_gradient(star)

        scaling_factor = const.G * system.primary.mass / system.semi_major_axis**2
        p_grad = calculate_potential_gradient(components_distance, component, points=points,
                                              synchronicity=synchronicity, mass_ratio=mass_ratio)
        g_acc_vector = scaling_factor * p_grad

        gravity = np.mean(np.linalg.norm(g_acc_vector, axis=1)[faces], axis=1) if star.symmetry_test else \
            np.mean(np.linalg.norm(g_acc_vector, axis=1), axis=1)
        setattr(star, 'potential_gradient_magnitudes', star.mirror_face_values(gravity)) \
            if star.symmetry_test() else setattr(star, 'potential_gradient_magnitudes', gravity)
        setattr(star, 'log_g', np.log10(star.potential_gradient_magnitudes))

        if star.has_spots():
            g_acc_vector_spot = dict()
            for spot_index, spot in star.spots.items():
                logger.debug(f'calculating surface SI unit gravity of {component} component / {spot_index} spot')
                logger.debug(f'calculating distribution of potential gradient '
                             f'magnitudes of spot index: {spot_index} / {component} component')

                p_grad = calculate_potential_gradient(components_distance, component, points=spot.points,
                                                      synchronicity=synchronicity, mass_ratio=mass_ratio)
                g_acc_vector_spot.update({spot_index: scaling_factor * p_grad})

                setattr(spot, 'potential_gradient_magnitudes',
                        np.mean(np.linalg.norm(g_acc_vector_spot[spot_index], axis=1)[spot.faces], axis=1))
                setattr(spot, 'log_g', np.log10(spot.potential_gradient_magnitudes))

    return system



import numpy as np

from abc import ABCMeta
from copy import copy, deepcopy
from . transform import (
    RVDataProperties,
    LCDataProperties
)
from .. dataset.graphic import plot
from .. dataset import utils as dutils
from ...base.types import INT
from ... logger import getLogger
from ... import utils, units as u
from ... import settings
from ... utils import is_empty

logger = getLogger('analytics.dataset.base')


class DataSet(metaclass=ABCMeta):
    """
    Abstract class dedicated to store any synthetic or (primarily) observational data.
    """
    TRANSFORM_PROPERTIES_CLS = None
    ID = 1

    def __init__(self, name=None, **kwargs):
        # initial kwargs
        self.kwargs = copy(kwargs)
        self.plot = plot.Plot(self)

        if is_empty(name):
            self.name = str(DataSet.ID)
            logger.debug(f"name of class instance {self.__class__.__name__} set to {self.name}")
            DataSet.ID += 1
        else:
            self.name = str(name)

        # initializing parmas to default values
        self.x_data = np.array([])
        self.x_unit = None
        self.y_data = np.array([])
        self.y_err = None

        self.check_data_validity(**kwargs)

    def transform_input(self, **kwargs):
        return self.__class__.TRANSFORM_PROPERTIES_CLS.transform_input(**kwargs)

    def copy(self):
        """
        Returns a deep copy of the Dataset.

        :return: Dataset;
        """
        return deepcopy(self)

    @staticmethod
    def check_data_validity(**kwargs):
        """
        Function makes sure that the data in the initialized data container are valid (i.e shapes are the same,
        not containing invalid numbers). Function chcecks for:

            - same dimensions of times stamps and observables (with their errors)
            - checking for invalid numbers inside the dataset

        :param kwargs: Dict;
        :**kwargs contents**:
            * :x_data: numpy.array; times of observations or photometric phases
            * :y_data: numpy.array; observables
            * :y_err: numpy.array; errors of observables

        :return:
        """
        if not np.shape(kwargs['x_data']) == np.shape(kwargs['y_data']):
            raise ValueError('`x_data` and `y_data` are not of the same shape.')
        if 'y_err' in kwargs.keys() and not utils.is_empty(kwargs['y_err']):
            if not np.shape(kwargs['x_data']) == np.shape(kwargs['y_err']):
                raise ValueError('`y_err` are not of the same shape to `x_data` and `y_data`.')

        # check for nans
        if np.isnan(kwargs['x_data']).any():
            raise ValueError('`x_data` contains NaN')
        if np.isnan(kwargs['y_data']).any():
            raise ValueError('`y_data` contains NaN')
        if 'y_err' in kwargs.keys() and not utils.is_empty(kwargs['y_err']):
            if np.isnan(kwargs['y_err']).any():
                raise ValueError('`y_err` contains NaN')

    @classmethod
    def load_from_file(cls, filename, x_unit=None, y_unit=None, data_columns=None,
                       delimiter=settings.DELIM_WHITESPACE, downselect_ratio=None, **kwargs):
        """
        Function loads a RV/LC measurements from a file.

        :param filename: str; name of the file
        :param x_unit: astropy.unit.Unit;
        :param y_unit: astropy.unit.Unit;
        :param data_columns: Tuple; ordered tuple with column indices of x_data, y_data, y_errors
        :param delimiter: str; regex to define columns separtor
        :param downselect_ratio: float; (0, 1), selects only a uniformly distributed portion of the data to reduce
                                        amount of data
        :param kwargs: Dict;
        :**kwargs options**:
            * :reference_magnitude: float; zero point for magnitude conversion in case of LCData

        :return: Union[RVData, LCData];
        """
        data_columns = (0, 1, 2) if data_columns is None else data_columns
        data = dutils.read_data_file(filename, data_columns, delimiter=delimiter)

        if downselect_ratio is not None:
            idxs = np.arange(0, data.shape[0], step=int(1.0/downselect_ratio), dtype=INT)
            data = data[idxs]

        try:
            errs = data[:, 2]
        except IndexError:
            errs = None
        return cls(x_data=data[:, 0],
                   y_data=data[:, 1],
                   y_err=errs,
                   x_unit=x_unit,
                   y_unit=y_unit,
                   **kwargs)

    from_file = load_from_file

    def convert_to_phases(self, period, t0, centre=0.0):
        """
        Function converts DataSet with x_data in time unit to dimensionless photometric phases according to an
        ephemeris.

        :param period: float; period according to which fold the data
        :param t0: float; reference time when phase = 0
        :param centre: float; phase curve will be centered around this phase
        """
        self.x_data = utils.jd_to_phase(self.x_data, period, t0, centre=centre)
        self.x_unit = u.dimensionless_unscaled

    def convert_to_time(self, period, t0, to_unit=u.DEFAULT_PERIOD_UNIT):
        """
        Function converts DataSet with x_data in dimensionless phases to time according to an ephemeris.

        :param period: float; period according to which fold the data
        :param t0: float; reference time when phase = 0
        :param to_unit: unit to assign the transformed data
        """
        self.x_data = self.x_data * period + t0
        self.x_unit = to_unit

    def smooth(self, method='central_moving_average', **kwargs):
        """
        Convenient function to perform smoothing of the phased data using various smoothing methods.

        :param method: str;
        :**method options**:

            * :`central_moving_average`: assign the average value to the centre of the bin defined by the
                                         number `n_bins` and radius `radius` in a number of bins

        :param kwargs: see different smoothing methods
        """
        available_methods = ['central_moving_average']
        if method == 'central_moving_average':
            n_bins = kwargs.get('n_bins', 100)
            radius = kwargs.get('radius', 2)
            cyclic_boundaries = kwargs.get('cyclic_boundaries', True)
            dutils.central_moving_average(self, n_bins=n_bins, radius=radius, cyclic_boundaries=cyclic_boundaries)
        else:
            raise NotImplementedError(f'Method {method} is not implemented. Try one of these: {available_methods}')


class RVData(DataSet):
    """
    Child class of elisa.analytics.dataset.base.Dataset class storing radial velocity measurement.
    RVData instance can be initialized in two different ways:

    - directly in code, using numpy arrays as x and y values, astropy.units to specify corresponding units and a
      reference magnitude of the object. Example:

      ::

        from elisa import RVData

        lc_data = RVData(x_data, x_unit, y_data, y_err, y_unit)
    
    - loading directly from a file:

      ::

        rv_data = RVData.load_from_file(
            filename,
            x_unit, y_unit,
            data_columns = (0, 1, 2)  # column indices containing x_data, y_data, y_err arrays
        )

    Input parameters:

    :param x_data: numpy.array; times of observation or photometric phases
    :param y_data: numpy.array; radial velocities
    :param y_err: numpy.array; radial velocity errors - optional
    :param x_unit: astropy.unit.Unit; if `None` or `astropy.unit.dimensionless_unscaled` is given,
                                      the `x_data are regarded as photometric phases, otherwise if unit is convertible
                                      to days, the `x_data` are regarded to be in JD
    :param y_unit: astropy.unit.Unit; velocity unit of the observed radial velocities and its errors
    """

    MANDATORY_KWARGS = settings.DATASET_MANDATORY_KWARGS
    OPTIONAL_KWARGS = settings.DATASET_OPTIONAL_KWARGS
    ALL_KWARGS = MANDATORY_KWARGS + OPTIONAL_KWARGS
    TRANSFORM_PROPERTIES_CLS = RVDataProperties

    __slots__ = ALL_KWARGS

    def __init__(self, name=None, **kwargs):
        utils.invalid_kwarg_checker(kwargs, self.__slots__, RVData)
        utils.check_missing_kwargs(self.MANDATORY_KWARGS, kwargs, instance_of=RVData)
        super().__init__(name, **kwargs)

        kwargs = self.transform_input(**kwargs)

        # conversion to base units
        kwargs = self.convert_arrays(**kwargs)
        self.check_data_validity(**kwargs)
        self.init_parameters(**kwargs)

    def init_parameters(self, **kwargs):
        logger.debug(f"initialising properties of class instance {self.__class__.__name__}")
        for kwarg in RVData.ALL_KWARGS:
            if kwarg in kwargs:
                setattr(self, kwarg, kwargs[kwarg])

    @staticmethod
    def convert_arrays(**kwargs):
        """
        Converting data and units to its base counterparts or keeping them dimensionless.

        :param kwargs: Dict;
        :**kwargs contents**:
            * :x_data: numpy.array; times of observations or photometric phases
            * :y_data: numpy.array; radial velocities RV
            * :y_err: numpy.array; RV errors
        :return: Dict; observations transformed into base units
        """
        # converting x-axis
        kwargs['x_data'] = dutils.convert_data(kwargs['x_data'], kwargs['x_unit'], u.DEFAULT_PERIOD_UNIT)
        kwargs['x_unit'] = dutils.convert_unit(kwargs['x_unit'], u.DEFAULT_PERIOD_UNIT)

        # converting y-axis
        kwargs['y_data'] = dutils.convert_data(kwargs['y_data'], kwargs['y_unit'], u.VELOCITY_UNIT)

        # convert errors
        if 'y_err' in kwargs.keys():
            kwargs['y_err'] = dutils.convert_data(kwargs['y_err'], kwargs['y_unit'], u.VELOCITY_UNIT)
        kwargs['y_unit'] = dutils.convert_unit(kwargs['y_unit'], u.VELOCITY_UNIT)

        return kwargs


class LCData(DataSet):
    """
        Child class of elisa.analytics.dataset.base.Dataset class storing light curves.
        LCData instance can be initialized in two different ways:

        - directly in code, using numpy arrays as x and y values, astropy.units to specify corresponding units and a
          reference magnitude of the object. Example:

          ::

            from elisa import LCData

            lc_data = LCData(x_data, x_unit, y_data, y_err, y_unit, reference_magnitude)

        - loading directly from a file:

          ::

            lc_data = LCData.load_from_file(
                filename,
                x_unit, y_unit,
                reference_magnitude,
                data_columns = (0, 1, 2)  # column indices containing x_data, y_data, y_err arrays
            )

        Input parameters:

        :param x_data: numpy.array; times of observation or photometric phases
        :param y_data: numpy.array; light curves supplied in normalized fluxes or magnitudes
        :param y_err: numpy.array; light curve errors - optional
        :param x_unit: astropy.unit.Unit; if `None` or `astropy.unit.dimensionless_unscaled` is given,
                                          the `x_data` are regarded as phases, otherwise if unit is convertible
                                          to days, the `x_data` are regarded to be in JD
        :param y_unit: astropy.unit.Unit; unit of the observed flux and its errors
        :param reference_magnitude: if the input light curve is in magnitudes, the user have to provide a reference
                                    magnitude to use during magnitude to relative flux conversion
    """
    MANDATORY_KWARGS = settings.DATASET_MANDATORY_KWARGS
    OPTIONAL_KWARGS = settings.DATASET_OPTIONAL_KWARGS + ['reference_magnitude', 'passband']
    ALL_KWARGS = MANDATORY_KWARGS + OPTIONAL_KWARGS
    TRANSFORM_PROPERTIES_CLS = LCDataProperties

    __slots__ = ALL_KWARGS

    def __init__(self, name=None, **kwargs):
        self.passband = None
        self.reference_magnitude = None

        utils.invalid_kwarg_checker(kwargs, self.__slots__, LCData)
        utils.check_missing_kwargs(self.MANDATORY_KWARGS, kwargs, instance_of=LCData)
        super().__init__(name, **kwargs)
        kwargs = self.transform_input(**kwargs)

        # conversion to base units
        kwargs = self.convert_arrays(**kwargs)
        self.check_data_validity(**kwargs)

        self.init_parameters(**kwargs)

    def init_parameters(self, **kwargs):
        logger.debug(f"initialising properties of class instance {self.__class__.__name__}")
        for kwarg in LCData.ALL_KWARGS:
            if kwarg in kwargs:
                setattr(self, kwarg, kwargs[kwarg])

    @staticmethod
    def convert_arrays(**kwargs):
        """
        Converting data and units to its base counterparts or keeping them dimensionless.

        :param kwargs: Dict;
        :**kwargs contents**:
            * :x_data: numpy.array; times of observations or photometric phases
            * :y_data: numpy.array; light curve
            * :y_err: numpy.array; light curve errors
        :return: Dict; observations transformed into base units
        """
        # converting x-axis
        kwargs['x_data'] = dutils.convert_data(kwargs['x_data'], kwargs['x_unit'], u.DEFAULT_PERIOD_UNIT)
        kwargs['x_unit'] = dutils.convert_unit(kwargs['x_unit'], u.DEFAULT_PERIOD_UNIT)
        kwargs['reference_magnitude'] = kwargs.get('reference_magnitude', None)

        # convert errors
        if 'y_err' in kwargs.keys():
            kwargs['y_err'] = dutils.convert_flux_error(kwargs['y_err'], kwargs['y_unit'],
                                                        zero_point=kwargs['reference_magnitude'])

        # converting y-axis
        kwargs['y_data'] = dutils.convert_flux(kwargs['y_data'], kwargs['y_unit'],
                                               zero_point=kwargs['reference_magnitude'])
        kwargs['y_unit'] = dutils.convert_unit(kwargs['y_unit'], u.dimensionless_unscaled)

        return kwargs



from .. import settings

if settings.CUDA:
    import cupy as cp

    class CupyTensor(cp.ndarray):
        def __new__(cls, input_array, dtype='float32'):
            obj = cp.asarray(input_array, dtype=dtype).view()
            return obj

    Tensor = CupyTensor

else:
    import numpy as np

    class NumpyTensor(np.ndarray):
        def __new__(cls, input_array, dtype='float32'):
            obj = np.asarray(input_array, dtype=dtype).view(cls)
            return obj

        def get(self):
            return np.asarray(self)

    Tensor = NumpyTensor

__all__ = 'Tensor',



import matplotlib.pyplot as plt
from .. import units as u


def display_observations(**kwargs):
    """
    Graphics method for displaying the DataSet content.

    :param kwargs: Dict;
    :**kwargs options**:
        * :x_data: numpy.array;
        * :x_unit: astropy.unit.Unit;
        * :y_data: numpy.array;
        * :y_err: numpy.array;
        * :y_unit: astropy.unit.Unit;
        * :plot_kwargs: Dict; plot arguments fot plotting functions
                              (matplotlib.pyplot.errorbar or matplotlib.pyplot.scatter).

    """
    if kwargs['y_err'] is not None:
        plt.errorbar(x=kwargs['x_data'], y=kwargs['y_data'], yerr=kwargs['y_err'], linestyle='none',
                     **kwargs['plot_kwargs'])
    else:
        plt.scatter(x=kwargs['x_data'], y=kwargs['y_data'], **kwargs['plot_kwargs'])

    x_lbl = 'Phase' if kwargs['x_unit'] == u.dimensionless_unscaled else f'Time [{kwargs["x_unit"]}]'
    y_lbl = 'Flux' if kwargs['y_unit'] == u.dimensionless_unscaled else f'Magnitude [{kwargs["y_unit"]}]'
    plt.xlabel(x_lbl)
    plt.ylabel(y_lbl)
    plt.subplots_adjust(top=0.98, right=0.98)

    plt.show()



import numpy as np

from . transform import OrbitProperties
from ... import (
    const as c,
    units as u,
    utils
)
from ... base.orbit.orbit import AbstractOrbit
from ... base.types import INT, FLOAT
from ... logger import getLogger

logger = getLogger('single_system.orbit.orbit')


def angular_velocity(rotation_period):
    """
    Rotational angular velocity of the star.

    :param rotation_period: float;
    :return: float;
    """
    return c.FULL_ARC / (rotation_period * u.DefaultSingleSystemUnits.system.rotation_period).to(u.TIME_UNIT).value


def true_phase_to_azimuth(phase):
    """
    Calculates observer azimuths for single star system.

    :param phase: Union[numpy.array, float];
    :return: Union[numpy.array, float];
    """
    return c.FULL_ARC * phase


def azimuth_to_true_phase(azimuth):
    """
    :param azimuth: Union[numpy.array, float];
    :return: Union[numpy.array, float];
    """
    return azimuth / c.FULL_ARC


class Orbit(AbstractOrbit):
    """
    Model which represents rotational motion of the single system as apparent orbital motion of the observer.

    :param rotational_period: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity];
    :param inclination: Union[float, astropy.units.Quantity]; If unitless values is supplied, default unit
                          suppose to be radians.
    """
    MANDATORY_KWARGS = ['rotation_period', 'inclination']
    OPTIONAL_KWARGS = ['phase_shift']
    ALL_KWARGS = MANDATORY_KWARGS + OPTIONAL_KWARGS

    def __init__(self, **kwargs):
        utils.invalid_kwarg_checker(kwargs, Orbit.ALL_KWARGS, Orbit)
        utils.check_missing_kwargs(self.__class__.MANDATORY_KWARGS, kwargs, instance_of=self.__class__)
        kwargs = OrbitProperties.transform_input(**kwargs)

        super(Orbit, self).__init__(**kwargs)

        # default valeus of properties
        self.rotational_period = np.nan

        # values of properties
        logger.debug(f"setting properties of orbit")
        for kwarg in kwargs:
            setattr(self, kwarg, kwargs[kwarg])

        self.period = self.rotational_period

    def orbital_motion(self, phase):
        return self.rotational_motion(phase)

    @staticmethod
    def rotational_motion(phase):
        """
        Function takes photometric phase of the single system as input and calculates azimuths for the observer

        :param phase: Union[numpy.array, float];
        :return: numpy.array; matrix consisting of column stacked vectors azimut, angle and phase, last 3 columns of
                              const.Position

        ::

            numpy.array((az1, np.nan, np.nan, phs1),
                        (az2, np.nan, np.nan, phs2),
                         ...
                        (azN, np.nan, np.nan, phsN))
        """
        # ability to accept scalar as input
        if isinstance(phase, (int, INT, float, FLOAT)):
            phase = np.array([FLOAT(phase)])

        azimuth_angle = true_phase_to_azimuth(phase=phase)

        return np.column_stack((azimuth_angle, np.full(phase.shape, np.nan), phase))

    @staticmethod
    def rotational_motion_from_azimuths(azimuth):
        """
        return rotational motion derived from known azimuths

        :param azimuth: Union[numpy.array, float];
        :return: numpy.array; matrix consisting of column stacked vectors distance,
                              azimut angle, true anomaly and phase

        ::

               numpy.array((az1, phs1),
                           (az2, phs2),
                            ...
                           (azN, phsN))

        """
        true_phase = azimuth_to_true_phase(azimuth)
        return np.column_stack((azimuth, np.full(azimuth.shape, np.nan), true_phase))



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject

Pre-release steps
-----------------

1.  make sure all newly added data files are listed in `setup.py` in variable
    `package_data`as well as in `MANIFEST.in` file

2.  make sure version information in following destinations is up to date:

    - `version.py` in function `get_version()`
    - `README.rst` in yellow version badge https://img.shields.io/badge/version-<VERSION>-yellow.svg
    - in `src/elisa/__init__.py`, variable `__version__`

3.  make sure that `CHANGELOG.rst` is up to date; content as well as release date and valid version

4.  make sure all newly added dependencies are listed in `requirements.txt` as well as in
    `setup.py` in variable `install_requires`

5.  make sure a latest docstsring documentation is generated and there is no error during Sphinx HTML build
    (for more comprehensive information take a look into `docs/README.rst`)

6.  make sure setup.cfg contains all supported Python versions

7.  commit changes

8.  create release branch with name `release/<VERSION>` and push changes

9.  make sure all unittests are running

10. create a tag::

    >> git tag -a v<VERSION> -m "version <VERSION>"
    >> git push origin --tags

11. build package with command::

    >> python setup.py sdist bdist_wheel

12. configure pypi repositories if necessary (use following configuration template with valid credentials)::

        # ~/.pypirc
        [distutils] # this tells distutils what package indexes you can push to
        index-servers =
            pypi

        [pypi]
        repository = https://upload.pypi.org/legacy/
        username = <username>
        password = <password>


13. release packages with following command (install twine if necessary with `pip install twine`)::

    >> twine upload dist/* -r pypi
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path
from version import get_version


here = path.dirname(__file__)

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = "For more information visit https://github.com/mikecokina/elisa/blob/master/README.rst"

setup(
    name='elisa',
    src_root='src',
    version=get_version(),

    description='Eclipsing Binary Modeling Software',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/mikecokina/elisa',

    # Author details
    author='Michal Cokina, Miroslav Fedurco',
    author_email='mikecokina@gmail.com, mirofedurco@gmail.com',

    # Choose your license
    license='GPLv2',

    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Astronomy',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        # 'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],

    # What does your project relate to?
    keywords='eclipsing binaries astronomy analysis analytics physic',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    # packages=find_packages(where='src', exclude=["single_system"]),
    packages=find_packages(where='src'),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'astropy>=4.0.1,<=5.3.4',
        'corner>=2.2.1',
        'emcee>=3.0.1,<=3.1.6',
        'jsonschema>=3.2.0',
        'matplotlib>=3.3.2,<=3.7.5',
        'numpy>=1.16.2,<=1.26.4',
        'packaging<=24.2',
        'pandas>=0.24.0,<=1.5.3',
        'python-dateutil>=2.6.1,<=2.8.1',
        'scipy>=1.0.0,<=1.11.4',
        'tqdm>=4.43.0',
        'parameterized>=0.7.4',
        'numba>=0.51.2',
        'requests>=2.26.0',
        'setuptools<=75.8.0',
        'wheel<=0.45.1',
    ],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': [],
        'test': ['coverage', 'parameterized>=0.7.4', 'pytest>=3.2.3,<=8.3.4'],
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={
        'elisa': [
            'passband/*',
            'conf/*',
            'conf/logging_schemas/*',
            'schema_registry/*',
            'schema_registry/schema_files/*',
            'data/*',
            'data/mesh_corrections/*'
        ],
    },

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    data_files=[],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
        ],
    },
)



import numpy as np

from elisa import utils


def correct_component_mesh(star, com, correction_factors):
    """
    Correcting the underestimation of the surface due to the discretization.

    :param correction_factors: numpy.array; (2*N) [discretization factor, correction factor],
    :param com: float; position of center of mass of the system
    :param star: elisa.base.container.StarContainer;
    :return: elisa.base.container.StarContainer;
    """
    com_vector = np.array([com, 0, 0])
    args = (star.discretization_factor, correction_factors)
    centered_points = utils.discretization_correction_factor(*args) * (star.points - com_vector[None, :])
    star.points = centered_points + com_vector[None, :]

    if star.has_spots():
        for spot in star.spots.values():
            args = (spot.discretization_factor, correction_factors)
            centered_points = utils.discretization_correction_factor(*args) * (spot.points - com_vector[None, :])
            spot.points = centered_points + com_vector[None, :]

    return star


def symmetry_point_reduction(array, base_symmetry_points_number):
    """
    Returns portion of the points (or other surface point related distribution) contained in the 1st quadrant (octant).
    This is used during utilization of surface symmetries to reduce number of surface points that code works with.

    :param array: numpy.array; surface points or point related surface distribution
    :param base_symmetry_points_number: int; number of first n surface points stored in StarContainer.points that are
                                             located on a symmetrical part of the surface. E.g.:

                                             ::

                                                StarContainer.points[:StarContainer.base_symmetry_points_number]

                                              will select surface points from one quarter (eighth) of the star in case
                                              of binary (single) star system
    :return: numpy.array; reduced form of an input array from the symmetrical part of the surface
    """
    return array[:base_symmetry_points_number]



import numpy as np

from abc import abstractmethod

from .. import const
from .. logger import getLogger
from copy import (
    deepcopy,
    copy
)
from .. import (
    utils,
    umpy as up
)
from . surface.mesh import symmetry_point_reduction
from . surface.faces import mirror_face_values, symmetry_face_reduction

logger = getLogger("base.container")


class PropertiesContainer(object):
    """
    General object for storing attributes of modelled objects.
    """
    def __init__(self, **kwargs):
        self.properties = kwargs
        for k, v in kwargs.items():
            setattr(self, k, v)

    def to_dict(self):
        return self.properties

    def __getitem__(self, item):
        return getattr(self, item)

    def __str__(self):
        return str(self.to_dict())


class StarPropertiesContainer(PropertiesContainer):
    """
    Object for handling attributes of Star objects.
    """
    pass


class SystemPropertiesContainer(PropertiesContainer):
    """
    Object for handling attributes of System objects.
    """
    pass


class PositionContainer(object):
    """
    General object for handling system models at a given time or photometric phase.
    """
    def __init__(self, position):
        self._flatten = False
        self._components = list()
        self.position = position
        self.inclination = np.nan
        self.period = np.nan

    @abstractmethod
    def build(self, *args, **kwargs):
        pass

    @abstractmethod
    def build_mesh(self, *args, **kwargs):
        pass

    @abstractmethod
    def build_faces(self, *args, **kwargs):
        pass

    @abstractmethod
    def build_surface_areas(self, *args, **kwargs):
        pass

    @abstractmethod
    def build_faces_orientation(self, *args, **kwargs):
        pass

    @abstractmethod
    def build_surface_gravity(self, *args, **kwargs):
        pass

    @abstractmethod
    def build_temperature_distribution(self, *args, **kwargs):
        pass

    def is_flat(self):
        return self._flatten

    def flat_it(self):
        """
        Function merges surface attributes for each system component which were previously assigned to
        component surface or spot attributes. Resulting attributes such as points, faces, temperatures. etc.
        are then stored in unified arrays as component attributes.

        :return: PositionContainer; flattened container
        """
        # naive implementation of idempotency
        if self._flatten:
            return self

        for component in self._components:
            star_container = getattr(self, component)
            if star_container.has_spots() or star_container.has_pulsations():
                star_container.flat_it()

        self._flatten = True
        return self

    def apply_rotation(self):
        """
        Rotate quantities defined in __PROPERTIES_TO_ROTATE__.
        Rotation is made in orbital plane (around z-axis) and inclination (y-axis) direction in respective order.
        Angle are defined in self.position and self.inclination.

        :return: elisa.base.PositionContainer;
        """
        __PROPERTIES_TO_ROTATE__ = ["points", "normals", "velocities", "face_centres"]

        for component in self._components:
            star_container = getattr(self, component)
            for prop in __PROPERTIES_TO_ROTATE__:
                self.rotate_property(star_container, prop)
        return self

    def rotate_property(self, container, prop):
        """
        Rotating property of StarContainer from co-rotating frame to observer's frame of reference.

        :param container: base.StarContainer;
        :param prop: str; name of the property (e.g. 'points')
        :return:
        """
        prop_value = getattr(container, prop)
        prop_value = utils.rotate_item(prop_value, self.position, self.inclination)
        setattr(container, prop, prop_value)

    def add_secular_velocity(self):
        """
        Addition of secular radial velocity of centre of mass to convert velocities to reference frame of observer
        """
        gamma = getattr(self, "gamma")
        for component in self._components:
            star = getattr(self, component)
            star.velocities[:, 0] += gamma
        return self

    def apply_darkside_filter(self):
        """
        Apply darkside filter on current position defined in container.
        Function iterates over components and assigns indices of visible points of star_component instance.

        :return: elisa.base.PositionContainer;
        """

        for component in self._components:
            star_container = getattr(self, component)
            cosines = getattr(star_container, "los_cosines")
            valid_indices = self.darkside_filter(cosines=cosines)
            setattr(star_container, "indices", valid_indices)
        return self

    def calculate_face_angles(self, line_of_sight):
        """
        Calculates angles between normals and line_of_sight vector for all components of the system.

        :param line_of_sight: np.array;
        """
        for component in self._components:
            star_container = getattr(self, component)
            normals = getattr(star_container, "normals")
            los_cosines = self.return_cosines(normals, line_of_sight=line_of_sight)
            setattr(star_container, "los_cosines", los_cosines)

    @staticmethod
    def return_cosines(normals, line_of_sight):
        """
        returns angles between normals and line_of_sight vector

        :param normals: numpy.array;
        :param line_of_sight: numpy.array;
        :return:
        """
        return utils.calculate_cos_theta_los_x(normals=normals) \
            if np.array((line_of_sight == const.LINE_OF_SIGHT)).all() \
            else utils.calculate_cos_theta(normals=normals, line_of_sight_vector=line_of_sight)

    @staticmethod
    def darkside_filter(cosines):
        """
        Return indices for visible faces defined by given normals.
        Function assumes that `line_of_sight` ([1, 0, 0]) and `normals` are already normalized to one.

        :param cosines: numpy.array;
        :return: numpy.array;
        """
        # todo: resolve self shadowing in case of W UMa, but probably not here
        # recovering indices of points on near-side (from the point of view of observer)
        return up.arange(np.shape(cosines)[0])[cosines > 0]

    def copy(self):
        """
        Return deepcopy of PositionContainer instance.
        :return: elisa.base.container.PositionContainer;
        """
        return deepcopy(self)


class StarContainer(object):
    """
    Container carrying non-static properties of Star object (properties which vary from phase to phase) and also
    all properties set during creation of the parent BinarySystem.
    Method `from_properties_container` or `from_star_instance` has to be used to create container properly.
    Following parameters are gathered from Star object

    'mass', 't_eff', 'synchronicity', 'albedo', 'discretization_factor', 'name', 'spots'
    'polar_radius', 'equatorial_radius', 'gravity_darkening', 'surface_potential', 'atmosphere',
    'pulsations', 'metallicity', 'polar_log_g', 'critical_surface_potential', 'side_radius'

    If you are experienced user, you can create instance directly by calling

    >>> from elisa.base.container import StarContainer
    >>> kwargs = {}
    >>> inst = StarContainer(**kwargs)

    and setup static parameter latter of not at all if not necessary for further use. However, the StarContainer
    instance is intialized within a given PositionContainer during its initialization.

    Bellow are optional input parameters of StarContainer.
    kwargs: Dict;
    :**kwargs options**:

        * **points** * -- numpy.array;
        * **normals** * -- numpy.array;
        * **indices** * -- numpy.array;
        * **faces** * -- numpy.array;
        * **temperatures** * -- numpy.array;
        * **log_g** * -- numpy.array;
        * **coverage** * -- numpy.array;
        * **face_centres** * -- numpy.array;
        * **metallicity** * -- float;
        * **areas** * -- numpy.array;
        * **potential_gradient_magnitudes** * -- numpy.array;
        * **ld_cfs** * -- numpy.array;
        * **normal_radiance** * -- numpy.array;
        * **los_cosines** * -- numpy.array;


    Output parameters (obtained by applying given methods upon container).

    :points: numpy.array;

    ::

        Numpy array of points that form surface of Body.
        Input dictionary has to be in shape::
            points = numpy.array([[x1 y1 z1],
                                  [x2 y2 z2],
                                   ...
                                  [xN yN zN]])

        where xi, yi, zi are cartesian coordinates of vertice i.

    :normals: numpy.array; Array containing normalised outward facing normals
              of corresponding faces with same index

    ::

            normals = numpy_array([[normal_x1, normal_y1, normal_z1],
                                   [normal_x2, normal_y2, normal_z2],
                                    ...
                                   [normal_xn, normal_yn, normal_zn]]

    :faces: numpy.array;

    ::

        Numpy array of triangles that will create surface of body.
        Triangles are stored as list of indices of points.

            numpy.array([[vertice_index_k, vertice_index_l, vertice_index_m]),
                         [...]),
                          ...
                         [...]])

    :temperatures: numpy.array; Array of temeratures of corresponding faces.

    ::
            numpy.array([t_eff1, ..., t_effn])

    :log_g: numpy.array; Array of log_g (gravity) of corresponding faces.

    ::

        numpy.array([log_g1, ..., log_gn])

    :coverage: numpy.array; surface area of each triangle visible to the observer
    :indices: numpy.array; Indices of visible faces when darkside_filter is applied.
    :face_centres: numpy.array; row-wise geometrical centres of each triangle
    :metallicity: float;
    :areas: numpy.array; surface areas corresponing to each surface triangle
    :potential_gradient_magnitudes: numpy.array;
    :inverse_point_symmetry_matrix: numpy.array; row-wise, set of vectors that map base symmetry quadrant points
                                                 (octant in case of single star) to all others quadrants, each row
                                                 contains indices of points in given quadrant which order corresponds to
                                                 the order of points in the 1st quadrant (octant)
    :base_symmetry_points_number: int; number of first n surface points stored in StarContainer.points that are located
                                       on a symmetrical part of the surface. E.g.:

                                        ::

                                            StarContainer.points[:StarContainer.base_symmetry_points_number]

                                       will select surface points from one quarter (eighth) of the star in case of
                                       binary (single) star system
    :base_symmetry_faces_number: int; number of first n triangles stored in StarContainer.faces located in the 1st
                                      quadrant (octant). E.g.:

                                       ::

                                            StarContainer.temperatures[:StarContainer.base_symmetry_faces_number]

                                      will return surface temperatures only from triangles int the 1st quadrant
                                      (octant) in case of binary (single) star system.
    :face_symmetry_vector: numpy.array; array that maps each surface triangle to the symmetrical surface triangle
                                        located in the 1st quadrant. Array contains indices between 0 and
                                        `base_symmetry_faces_number`.
    :base_symmetry_points: numpy.array;
    :base_symmetry_faces: numpy.array;
    :polar_potential_gradient_magnitude: numpy.array;
    :ld_cfs: numpy.array;
    :normal_radiance: numpy.array;
    :los_cosines: numpy.array;
    """

    def __init__(self,
                 points=None,
                 normals=None,
                 velocities=None,
                 accelerations=None,
                 indices=None,
                 faces=None,
                 temperatures=None,
                 log_g=None,
                 coverage=None,
                 face_centres=None,
                 metallicity=None,
                 areas=None,
                 potential_gradient_magnitudes=None,
                 ld_cfs=None,
                 normal_radiance=None,
                 los_cosines=None):

        self.points = points
        self.normals = normals
        self.faces = faces
        self.velocities = velocities
        self.accelerations = accelerations
        self.temperatures = temperatures
        self.log_g = log_g
        self.coverage = coverage
        self.indices = indices
        self.face_centres = face_centres
        self.metallicity = metallicity
        self.areas = areas
        self.potential_gradient_magnitudes = potential_gradient_magnitudes
        self.ld_cfs = ld_cfs
        self.normal_radiance = normal_radiance
        self.los_cosines = los_cosines
        self.points_spherical = np.array([])
        self.com = np.array([])

        self.inverse_point_symmetry_matrix = np.array([])
        self.base_symmetry_points_number = 0

        self.face_symmetry_vector = np.array([])
        self.base_symmetry_faces_number = 0

        # aux variables for treating singularities at poles
        self.pole_idx = np.array([])
        self.pole_idx_neighbour = np.array([])

        # those are used only if case of spots are NOT used ------------------------------------------------------------
        self.base_symmetry_points = np.array([])
        self.base_symmetry_faces = np.array([])
        self.azimuth_args = np.array([])
        # --------------------------------------------------------------------------------------------------------------

        self.spots = dict()
        self.pulsations = dict()
        self.polar_potential_gradient_magnitude = np.nan

        # all star radii in any position (set on fly) ------------------------------------------------------------------
        # set only via `assign_radii()` method
        self.polar_radius = None
        self.forward_radius = None
        self.side_radius = None
        self.backward_radius = None
        self.equatorial_radius = None
        self.equivalent_radius = None
        # --------------------------------------------------------------------------------------------------------------

        self._flatten = False

    @classmethod
    def from_star_instance(cls, star):
        """
        Initialize StarContainer from Star instance.

        :param star: elisa.base.star.Star;
        :return: StarContainer;
        """
        return cls.from_properties_container(star.to_properties_container())

    @classmethod
    def from_properties_container(cls, properties_container):
        """
        Create StarContainer from properties container.

        :param properties_container: elisa.base.container.StarPropertiesContainer;
        :return: elisa.base.container.StarContainer;
        """
        container = cls()
        container.__dict__.update(properties_container.__dict__)
        return container

    def has_spots(self):
        """
        True if StarContainer contains spots.

        :return: bool;
        """
        return len(self.spots) > 0

    def has_pulsations(self):
        """
        True if StarContainer contains pulsation modes.

        :return: bool;
        """
        return len(self.pulsations) > 0

    def symmetry_test(self):
        """
        True if stellar surface can utilize surface symmetry during the model calculation.

        :return: bool;
        """
        return not self.has_spots() and not self.has_pulsations()

    def is_flat(self):
        """
        Returns True if StarContainer is flattened.

        :return: bool;
        """
        return self._flatten

    def copy(self):
        """
        Return deepcopy of StarContainer instance.

        :return: elisa.base.container.StarContainer;
        """
        return deepcopy(self)

    def remove_spot(self, spot_index: int):
        """
        Remove n-th spot index of object.

        :param spot_index: int
        """
        del (self.spots[spot_index])

    def get_flatten_points_map(self):
        """
        Function returns all surface point and faces with corresponding map of vertices.

        :param self: Star instance
        :return: Tuple[numpy.array, Any]: [all surface points including star and surface points, vertices map or None]
        """
        points = copy(self.points)
        for spot_index, spot_instance in self.spots.items():
            points = up.concatenate([points, spot_instance.points])

        vertices_map = [{"type": "object", "enum": -1}] * len(self.points)
        for spot_index, spot_instance in self.spots.items():
            vertices_map = up.concatenate(
                [vertices_map, [{"type": "spot", "enum": spot_index}] * len(spot_instance.points)]
            )
        return points, vertices_map

    def calculate_areas(self):
        """
        Returns areas of each face of the star surface. (spots not included if StarContainer is not flattened)

        :return: numpy.array:

        ::

            numpy.array([area_1, ..., area_n])
        """
        if len(self.faces) == 0 or len(self.points) == 0:
            raise ValueError('Faces or/and points of object {self.name} have not been set yet.\n'
                             'Run build method first.')
        if self.symmetry_test():
            base_areas = utils.triangle_areas(self.symmetry_faces(self.faces), self.symmetry_points())
            return self.mirror_face_values(base_areas)
        return utils.triangle_areas(self.faces, self.points)

    def calculate_all_areas(self):
        """
        Calculates areas for all faces on the surface including spots and assigns values to its corresponding variables.
        """
        self.areas = self.calculate_areas()
        if self.has_spots() and not self.is_flat():
            for spot_index, spot_instance in self.spots.items():
                spot_instance.areas = spot_instance.calculate_areas()

    def surface_serializer(self):
        """
        Returns all points and faces of the whole star.

        :return: Tuple[numpy.array, numpy.array]
        """
        points = copy(self.points)
        faces = copy(self.faces)
        if self.has_spots():
            for spot_index, spot in self.spots.items():
                n_points = np.shape(points)[0]
                points = np.append(points, spot.points, axis=0)
                faces = np.append(faces, spot.faces + n_points, axis=0)

        return points, faces

    def reset_spots_properties(self):
        """
        Reset computed spots properties
        """
        for _, spot_instance in self.spots.items():
            spot_instance.boundary = np.array([])
            spot_instance.boundary_center = np.array([])
            spot_instance.center = np.array([])

            spot_instance.points = np.array([])
            spot_instance.normals = np.array([])
            spot_instance.faces = np.array([])
            spot_instance.face_centres = np.array([])

            spot_instance.areas = np.array([])
            spot_instance.potential_gradient_magnitudes = np.array([])
            spot_instance.temperatures = np.array([])
            spot_instance.velocities = np.array([])

            spot_instance.log_g = np.array([])

    def get_flatten_parameter(self, prop):
        """
        Returns flattened property of the container.

        :param prop: str; property identifier (e.g. 'points')
        :return: numpy.array; flattened property
        """
        list_to_concat = [getattr(self, prop), ]
        if not self.has_spots() or self._flatten:
            return list_to_concat[0]

        list_to_concat += [getattr(spot, prop) for spot in self.spots.values()]
        if prop == 'faces':
            lengths = np.cumsum([np.max(item) + 1 for item in list_to_concat])
            list_to_concat[1:] = [list_to_concat[index+1] + lengths[index] for index in self.spots]
        return np.concatenate(list_to_concat, axis=0)

    def flat_it(self):
        """
        Make properties "points", "normals", "faces", "temperatures", "log_g", "centers", "areas"
        of container flat. It means all properties of start and spots are put together.

        :return: self
        """
        # naive implementation of idempotency
        if self._flatten:
            return self

        props_list = ["points", "normals", "faces", "temperatures", "log_g", "face_centres", "areas", "velocities"]

        for prop in props_list:
            setattr(self, prop, self.get_flatten_parameter(prop))

        self._flatten = True
        return self

    def transform_points_to_spherical_coordinates(self, kind='points', com_x=0):
        """
        Transforming container cartesian variable to spherical coordinates in a frame of reference of given object.

        :param kind: str; `points` or `face_centres` or other variable containing cartesian
                          points in both star and spot containers
        :param com_x: float;
        :return: numpy.array; spherical coordinates of star variable
        """
        # separating variables to convert
        centres_cartesian = copy(getattr(self, kind))

        # transforming variables
        centres_cartesian[:, 0] -= com_x

        # conversion
        centres = utils.cartesian_to_spherical(centres_cartesian)
        return centres

    def assign_radii(self, radii):
        """
        Function automatically assigns radii stored in radii dict.

        :param radii: Dict; {radius_type: value, ...}
        """
        for key, value in radii.items():
            setattr(self, key, value)

    def symmetry_points(self):
        """
        Returns surface points from the symmetry base part of the star.

        :return: numpy.array;
        """
        # pulsations are not yet introduced during the creation of point cloud
        if self.has_spots():
            raise ValueError('Surface symmetry is not applicable in this case')
        return symmetry_point_reduction(self.points, self.base_symmetry_points_number)

    def mirror_face_values(self, values):
        """
        Mirroring the array values on symmetrical faces to the rest of the surface.

        :param values: numpy.array; surface parameter values corresponding to the base symmetry part of the surface
        :return: numpy.array; remapped array
        """
        if not self.symmetry_test():
            raise ValueError('Surface symmetry is not applicable in this case.')
        return mirror_face_values(values, self.face_symmetry_vector)

    def symmetry_faces(self, values):
        """
        Reducing the surface distribution of surface `values` array to a symmetrical component.

        :param values: numpy.array; surface parameter distribution
        :return: numpy.array; reduced surface distribution array
        """
        if not self.symmetry_test():
            raise ValueError('Surface symmetry is not applicable in this case.')
        return symmetry_face_reduction(values, self.base_symmetry_faces_number)




try:
    # noinspection PyProtectedMember
    from collections import Iterable
except ImportError:
    from typing import Iterable

import numpy as np

from abc import ABCMeta, abstractmethod
from ... pypex.poly2d.point import Point


class Shape2D(metaclass=ABCMeta):
    """
    Convex 2D Shape
    """

    def __init__(self, hull, _validity=True):
        hull = self.normalize_hull(hull)
        if _validity:
            self.polygon_validity_check(hull=hull, _raise=True)
        self._hull = np.array(hull)
        self.xi, self.yi = 0, 1

    def __len__(self):
        return len(self.hull)

    def __eq__(self, other):
        if len(other) != len(self):
            return False
        return all([Point(*self.hull[i]) == Point(*other.hull[i]) for i in range(len(self))])

    def __str__(self):
        return "Poly ({}): [{}]".format(len(self.hull), ", ".join([str(v) for v in self.hull]))

    def __repr__(self):
        return "Poly ({}): [{}]".format(len(self.hull), ", ".join([str(v) for v in self.hull]))

    @staticmethod
    def normalize_hull(hull):
        """
        Convert definition of points to  normal form (to numpy.array).

        :param hull: Iterable; iterable of points
        :return: numpy.array
        """
        return np.array([vertex.to_array() if isinstance(vertex, Point) else vertex for vertex in hull])

    @property
    def hull(self):
        return self._hull

    @hull.setter
    def hull(self, hull):
        self._hull = np.array(hull)

    @abstractmethod
    def intersects(self, shape, **kwargs):
        pass

    @abstractmethod
    def intersection(self, shape):
        pass

    def sort_clockwise(self, inplace=False):
        """
        Sort points of convex polygon in clokwise order

        :param inplace: bool; replace current polygon hull with new obtained values
        :return: numpy.array
        """
        center = np.sum(self.hull, axis=0) / self.hull.shape[0]
        x, y = self.hull.T[self.xi] - center[self.xi], self.hull.T[self.yi] - center[self.yi]
        atan2 = np.arctan2(y, x)
        arr1inds = atan2.argsort()[::-1][:len(atan2)]
        hull = self.hull[arr1inds[::-1]]
        if inplace:
            self.hull = hull
        return hull

    def to_Points(self):
        return np.array([Point(*point) for point in self.hull])

    def to_array(self):
        return np.array(self._hull)

    @staticmethod
    def validity_check(hull, length, _raise=True):
        length_test = (len(hull) == length) if length in [3, 2] else (len(hull) > length)
        try:
            if length_test & (isinstance(hull, (Iterable, np.ndarray))) & np.all(np.array([len(v) == 2 for v in hull])):
                return True
        except TypeError:
            pass
        if _raise:
            raise ValueError("invalid 2D polygon shape")
        return False

    @classmethod
    def polygon_validity_check(cls, hull, _raise=True):
        return cls.validity_check(hull, 1, _raise)

    @property
    def transpose(self):
        """
        Transpose self._hull and return numpy.array.
        :return: numpy.array
        """
        return self._hull.T

    T = transpose



from numba import jit
import numpy as np


@jit(nopython=True, cache=True)
def gamma_primary(normals, join_vector):
    """
    Calculate cosine of angles between visible normals of primary surface elements and each counterpart element on the
    secondary.

    :param normals: numpy.array; (a, 3)
    :param join_vector: numpy.array; (a, b, 3)
    :return: numpy.array; (a, b)
    """
    result = np.empty(join_vector.shape[:-1])
    for ii in range(normals.shape[0]):
        for jj in range(join_vector.shape[1]):
            result[ii, jj] = normals[ii, 0] * join_vector[ii, jj, 0] + \
                             normals[ii, 1] * join_vector[ii, jj, 1] + \
                             normals[ii, 2] * join_vector[ii, jj, 2]

    return result


@jit(nopython=True, cache=True)
def gamma_secondary(normals, join_vector):
    """
    Calculate cosine of angles between visible normals of primary surface elements and each counterpart element on the
    secondary.

    :param normals: numpy.array; (b, 3)
    :param join_vector: numpy.array; (a, b, 3)
    :return: numpy.array; (a, b)
    """
    result = np.empty(join_vector.shape[:-1])
    for ii in range(join_vector.shape[0]):
        for jj in range(join_vector.shape[1]):
            result[ii, jj] = - normals[jj, 0] * join_vector[ii, jj, 0] - \
                             normals[jj, 1] * join_vector[ii, jj, 1] - \
                             normals[jj, 2] * join_vector[ii, jj, 2]

    return result




import itertools
import json
import os
import sys
import warnings
import numpy as np
import pandas as pd

from queue import Queue
from threading import Thread
from typing import Iterable
from copy import deepcopy

from .base.types import FLOAT
from .logger import getLogger
from .base.error import (
    AtmosphereError,
    MetallicityError,
    TemperatureError,
    GravityError, ElisaError)
from scipy import (
    integrate,
    interpolate
)
from . import settings
from . import (
    umpy as up,
    utils,
    const,
    ld,
)
from . buffer import buffer
from .tensor.etensor import Tensor

logger = getLogger(__name__)


# * 1e-7 * 1e4 * 1e10 * (1.0/const.PI)

class AtmModel(object):
    def __init__(self, flux, wavelength):
        self.flux: np.array = flux
        self.wavelength: np.array = wavelength

    def _empty(self):
        return len(self.wavelength) == 0

    @property
    def empty(self):
        return self._empty()

    @classmethod
    def from_dataframe(cls, df):
        return cls(flux=np.array(df[settings.ATM_MODEL_DATAFRAME_FLUX], dtype=FLOAT),
                   wavelength=np.array(df[settings.ATM_MODEL_DATAFRAME_WAVE], dtype=FLOAT))

    def to_dataframe(self):
        return pd.DataFrame(
            {
                settings.ATM_MODEL_DATAFRAME_FLUX: self.flux,
                settings.ATM_MODEL_DATAFRAME_WAVE: self.wavelength
            }
        )

    def last_valid_index(self):
        return len(self.flux)

    def __getitem__(self, item):
        return AtmModel(flux=self.flux[item], wavelength=self.wavelength[item])

    def __len__(self):
        return len(self.wavelength)


class AtmDataContainer(object):
    def __init__(self, model, temperature, log_g, metallicity, fpath=''):
        self._model = AtmModel(flux=None, wavelength=None)
        self.temperature = temperature
        self.log_g = log_g
        self.metallicity = metallicity
        self.flux_unit = "flam"
        self.wave_unit = "angstrom"
        # in case this const.PI will stay here, there will be rendundant multiplication in intensity integration
        # flam = erg * s-1* cm-2 *A-1 = (10-7 * J) * s-1 * (10-2 * m)-2 * (10-10 * m)-1 =
        #        10-7 * 10**4 * 10**10 * J * s-1 * m-3
        self.flux_to_si_mult = 1e7  # * (1.0/const.PI)
        self.wave_to_si_mult = 1e-10
        self.left_bandwidth = np.nan
        self.right_bandwidth = np.nan
        self.fpath = fpath

        setattr(self, 'model', model)

    def is_empty(self):
        """
        Find out wheter model container which carries of atmospheric
        model AmtDatContainer instance is empty.

        :return: bool;
        """
        return self._model.empty

    @property
    def model(self):
        """
        Return atmospheric model instance.

        :return: elisa.atm.AtmModel;
        """
        return self._model

    @model.setter
    def model(self, data):
        """
        Setup model container which carries DataFrame of atmospheric model
        and left and right bandwidth of such container as well.

        :param data: Union[pandasDataFrame, elisa.atm.AtmModel];
        """
        self._model = AtmModel.from_dataframe(data) if isinstance(data, pd.DataFrame) else data
        self.left_bandwidth = self._model.wavelength.min()
        self.right_bandwidth = self._model.wavelength.max()


class IntensityContainer(object):
    """
    Intended to keep information about integrated radiance for given params.
    """

    def __init__(self, intensity, temperature, log_g, metallicity):
        """
        Initialise container with given parametres.

        :param intensity: float; integrated radiance
        :param temperature: float;
        :param log_g: float;
        :param metallicity: float;
        """
        self.intensity = intensity
        self.temperature = temperature
        self.log_g = log_g
        self.metallicity = metallicity


class NaiveInterpolatedAtm(object):
    @staticmethod
    def radiance(temperature, log_g, metallicity, atlas, **kwargs):
        """
        Compute radiance for given atmospheric parametres and given passbands.

        :param temperature: numpy.array;
        :param log_g: numpy.array;
        :param metallicity: float;
        :param atlas: str;
        :param kwargs:
        :**kwargs options**:
            * **left_bandwidth** * -- float; maximal allowed wavelength from left (Angstrom)
            * **right_bandwidth** * -- float; maximal allowed wavelength from right (Angstrom)
            * **passband** * -- Dict[str, elisa.observer.observer.PassbandContainer]
        :return: List;
        """
        if validated_atlas(atlas) == "bb":
            return NaiveInterpolatedAtm.black_body_radiance(temperature, **kwargs)
        return NaiveInterpolatedAtm.atlas_radiance(temperature, log_g, metallicity, atlas, **kwargs)

    @staticmethod
    def black_body_radiance(temperature, **kwargs):
        """
        Compute integrated flux based on values obtained from Planck Function (for purpose of black_body atmosphere).

        :param temperature: numpy.array;
        :param kwargs: Dict;
        :**kwargs options**:
            * **passband** * -- Dict[str, elisa.observer.observer.PassbandContainer]
        :return: Dict[str, float];
        """
        # setup multiplicators to convert quantities to SI
        flux_mult, wave_mult = const.PI, 1e-10
        # obtain localized atmospheres in matrix
        localized_atms = NaiveInterpolatedAtm.arange_black_body_localized_atms(temperature, kwargs["passband"])
        # integrate flux
        return compute_normal_radiances(localized_atms, flux_mult=flux_mult, wave_mult=wave_mult)

    @staticmethod
    def arange_black_body_localized_atms(temperature, passband_containers):
        """
        The function generates atmosphere models based on Planck Function.
        The all models are sitting on temperatures given by surface elements.

        :param temperature: numpy.array;
        :param passband_containers: elisa.observer.observer.PassbandContainer;
        :return: Dict[str, numpy.array];
        """
        localized_atms = dict()
        standard_wavelength = get_standard_wavelengths()

        # build temperature mask and avoid repeative computation
        # temperature values where decimal points are basicaly useless
        temperature = np.round(temperature, 0)
        temperature, reverse_map = np.unique(temperature, return_inverse=True)

        for band, pb_container in passband_containers.items():
            # how many wavelengths generate based on standard
            mask = np.logical_and(np.less_equal(standard_wavelength, pb_container.right_bandwidth),
                                  np.greater_equal(standard_wavelength, pb_container.left_bandwidth))
            hm_waves = len(standard_wavelength[mask])
            # wavelenghts in angstrom
            wavelengths = np.sort(np.unique(
                np.concatenate(
                    [np.linspace(pb_container.left_bandwidth, pb_container.right_bandwidth, hm_waves, True),
                     standard_wavelength[mask]])
            ))

            # compute flux in flam, apply passband and replace possible NaNs
            flux = np.nan_to_num([
                pb_container.akima(wavelengths) *
                planck_function(wavelengths * pb_container.wave_to_si_mult, _temperature)
                for _temperature in temperature
            ])
            # sometimes, there are small negative values on the boundwidth boundaries due to akima interpolation
            flux[np.less(flux, 0)] = 0.0
            # broadcast and fill localized atms
            localized_atms[band] = {"flux": flux[reverse_map], "wave": wavelengths}

        return localized_atms

    @staticmethod
    def get_atm_profiles(temperature, log_g, metallicity, atlas, **kwargs):
        """
        Returns atmosphere profiles for given surface parameters.

        :param temperature: Iterable[float];
        :param log_g: Iterable[float];
        :param metallicity: float;
        :param atlas: str; atmosphere model identificator (see settings.ATLAS_TO_ATM_FILE_PREFIX.keys())
        :param kwargs: Dict;
        :return: Tuple[Dict, numpy.float, numpy.float]; atmosphere profiles for each passband, flux multiplicator,
                                                        wave multiplicator;
        """
        l_bandw, r_bandw = kwargs["left_bandwidth"], kwargs["right_bandwidth"]
        passband_containers = kwargs["passband"]
        # related atmospheric files for each face (upper and lower)
        atm_files = NaiveInterpolatedAtm.atm_files(temperature, log_g, metallicity, atlas)
        # find unique atmosphere data files
        unique_atms, containers_map = read_unique_atm_tables(atm_files)
        # get multiplicators to transform containers from any units to si
        flux_mult, wave_mult = find_atm_si_multiplicators(unique_atms)
        # common wavelength coverage of atmosphere models
        # intersection of wavelengths of models
        global_left, global_right = find_global_atm_bandwidth(unique_atms)
        # strip unique atmospheres to passbands coverage
        unique_atms = strip_atm_containers_by_bandwidth(unique_atms, l_bandw, r_bandw,
                                                        global_left=global_left, global_right=global_right)

        # alignement of atmosphere containers wavelengths for convenience
        unique_atms = arange_atm_to_same_wavelength(unique_atms)
        passbanded_atm_containers = apply_passband(unique_atms, passband_containers,
                                                   global_left=global_left, global_right=global_right)

        flux_matrices = remap_passbanded_unique_atms_to_matrix(passbanded_atm_containers, containers_map)
        atm_containers = remap_passbanded_unique_atms_to_origin(passbanded_atm_containers, containers_map)
        localized_atms = NaiveInterpolatedAtm.interpolate_spectra(atm_containers, flux_matrices,
                                                                  temperature=temperature)

        return localized_atms, flux_mult, wave_mult

    @staticmethod
    def atlas_radiance(temperature, log_g, metallicity, atlas, **kwargs):
        """
        Returns normal radiance for given surface parameters.

        :param temperature: Iterable[float];
        :param log_g: Iterable[float];
        :param metallicity: float;
        :param atlas: str; atmosphere model identificator (see settings.ATLAS_TO_ATM_FILE_PREFIX.keys())
        :param kwargs:
        :return: Dict;
        """
        args = temperature, log_g, metallicity, atlas
        localized_atms, flux_mult, wave_mult = NaiveInterpolatedAtm.get_atm_profiles(*args, **kwargs)
        return compute_normal_radiances(localized_atms, flux_mult=flux_mult, wave_mult=wave_mult)

    @staticmethod
    def compute_interpolation_weights(temperatures, top_atm_containers, bottom_atm_containers):
        """
        Compute interpolation weights between two models of atmoshperes.
        Weights are computet as::

            (temperatures^4 - bottom_temperatures^4) / (top_temperatures^4 - bottom_temperatures^4)

        what means we use linear approach.
        If there is np.nan (it cames from same surounded values), such value is replaced with 1.0.
        1.0 is choosen to fit interpolation method and return correct atmosphere.

        :param temperatures: numpy.array[float];
        :param top_atm_containers: numpy.array[elisa.atm.AtmDataContainer];
        :param bottom_atm_containers: numpy.array[elisa.atm.AtmDataContainer];
        :return: numpy.array[float];
        """
        with np.errstate(divide='ignore', invalid='ignore'):
            top_temperatures4 = np.power(np.array([a.temperature for a in top_atm_containers]), 4)
            bottom_temperatures4 = np.power(np.array([a.temperature for a in bottom_atm_containers]), 4)

            result = (np.power(temperatures, 4) - bottom_temperatures4) / (top_temperatures4 - bottom_temperatures4)

            result[up.isnan(result)] = 1.0
            return result

    @staticmethod
    def compute_unknown_intensity_from_surounded_containers(weight, top_atm_container, bottom_atm_container):
        """
        Depends on weight will compute (interpolate) intensities from surounded intensities
        related to given temperature.
        ! Top and bottom atmosphere model are have to be defined in same wavelengths !

        :param weight: Iterable[float];
        :param top_atm_container: elisa.atm.AtmDataContainer;
        :param bottom_atm_container: elisa.atm.AtmDataContainer;
        :return: Tuple[numpy.array, numpy.array]; (flux, wave);
        """
        if bottom_atm_container is None:
            return top_atm_container.model.flux, top_atm_container.model.wavelength

        # reset index is neccessary; otherwise add/mult/... method od DataFrame
        # leads to nan if left and right frame differ in indices
        top_atm_container.model.reset_index(drop=True, inplace=True)
        bottom_atm_container.model.reset_index(drop=True, inplace=True)

        intensity = weight * (
                top_atm_container.model.flux - bottom_atm_container.model.flux) + bottom_atm_container.model.flux

        return intensity, top_atm_container.model.wavelength

    @staticmethod
    def compute_unknown_intensity_from_surounded_flux_matrices(weights, top_flux_matrix, bottom_flux_matrix):
        import time
        t = time.time()

        weights = Tensor(weights)
        top_flux_matrix = Tensor(top_flux_matrix)
        bottom_flux_matrix = Tensor(bottom_flux_matrix)
        result = (weights * (top_flux_matrix.T - bottom_flux_matrix.T) + bottom_flux_matrix.T).T

        logger.info(f"time {time.time() - t}")
        return result.get()

    @staticmethod
    def interpolate_spectra(passbanded_atm_containers, flux_matrices, temperature):
        """
        From supplied elisa.atm.AtmDataContainer's, `flux_matrices` and `temeprature`.
        Interpolation is computed in vector form::

            (weights * (top_flux_matrix.T - bottom_flux_matrix.T) + bottom_flux_matrix.T).T

        where `top_flux_matrix` and `bottom_flux_matrix`, are entire matrix where rows are represented by fluxes.
        It also means, to be able do such interpolation, fluxes have to be on same wavelengths for each row.

        :param flux_matrices: Dict[str, numpy.array];

        ::

            {"passband": numpy.array (matrix)}

        :param passbanded_atm_containers: Dict[str, elisa.atm.AtmDataContainers];
        :param temperature: numpy.array[float];
        :return: Dict[str, numpy.array];
        """

        interp_band = dict()
        for band, flux_matrix in flux_matrices.items():
            band_atm = passbanded_atm_containers[band]
            bottom_flux, top_flux = flux_matrix[:len(flux_matrix) // 2], flux_matrix[len(flux_matrix) // 2:]
            bottom_atm, top_atm = band_atm[:len(band_atm) // 2], band_atm[len(band_atm) // 2:]

            logger.debug(f"computing atmosphere interpolation weights for band: {band}")
            interpolation_weights = NaiveInterpolatedAtm.compute_interpolation_weights(temperature, top_atm, bottom_atm)
            flux = NaiveInterpolatedAtm.compute_unknown_intensity_from_surounded_flux_matrices(
                interpolation_weights, top_flux, bottom_flux
            )
            interp_band[band] = {
                settings.ATM_MODEL_DATAFRAME_FLUX: flux,
                settings.ATM_MODEL_DATAFRAME_WAVE: find_atm_defined_wavelength(top_atm)
            }
        return interp_band

    @staticmethod
    def atm_files(temperature, log_g, metallicity, atlas):
        """
        Find out related atmospheric csv tables and return list of paths to them.

        :param temperature: Iterable[float];
        :param log_g: Iterable[float];
        :param metallicity: float;
        :param atlas: str; atmosphere model identificator (see settings.ATLAS_TO_ATM_FILE_PREFIX.keys())
        :return: List[str];
        """
        atlas = validated_atlas(atlas)
        log_g = utils.convert_gravity_acceleration_array(log_g, "log_cgs")

        g_array = np.array(atm_file_prefix_to_quantity_list("gravity", atlas))
        m_array = np.array(atm_file_prefix_to_quantity_list("metallicity", atlas))
        t_array = np.array(atm_file_prefix_to_quantity_list("temperature", atlas))

        g = utils.find_nearest_value_as_matrix(g_array, log_g)[0]
        m = utils.find_nearest_value_as_matrix(m_array, metallicity)[0][0]
        t = utils.find_surrounded_as_matrix(t_array, temperature)

        domain_df = pd.DataFrame({
            "temp": t.flatten('F'),
            "log_g": np.tile(g, 2),
            "mh": np.repeat(m, len(g) * 2)
        })
        directory = get_atm_directory(m, atlas)
        mh_name = domain_df["mh"].apply(lambda x: utils.numeric_metallicity_to_string(x))
        temp_name = domain_df["temp"].apply(lambda x: str(int(x)))
        log_g_name = domain_df["log_g"].apply(lambda x: utils.numeric_logg_to_string(x))
        fnames = str(atlas) + mh_name + "_" + temp_name + "_" + log_g_name

        return list(
            os.path.join(str(settings.ATLAS_TO_BASE_DIR[atlas]), str(directory)) + os.path.sep + fnames + ".csv"
        )


def arange_atm_to_same_wavelength(atm_containers):
    """
    Function aligns all atmosphere profiles to the same wavelengths.

    :param atm_containers: Iterable[elisa.atm.AtmDataContainer]; atmosphere containers which
                           wavelengths should be aligned
    :return: Iterable[elisa.atm.AtmDataContainer]; wavelength aligned atmospheric containers
    """

    wavelengths = np.unique(np.array([atm.model.wavelength for atm in atm_containers]).flatten())
    wavelengths.sort()
    result = list()

    # this code checks if the containers are already alligned
    s_size = sys.maxsize
    for atm in atm_containers:
        s_size = len(atm.model) if len(atm.model) < s_size else s_size

    # if yes, interpolation is unnecessary
    if s_size == len(wavelengths):
        return atm_containers

    # otherwise interpolation is utilized
    for atm in atm_containers:
        i = interpolate.Akima1DInterpolator(atm.model.wavelength, atm.model.flux)
        atm.model = AtmModel(wavelength=wavelengths, flux=np.nan_to_num(i(wavelengths)))
        result.append(atm)
    return result


def strip_atm_containers_by_bandwidth(atm_containers, left_bandwidth, right_bandwidth, **kwargs):
    """
    Strip all loaded atm models to common wavelength coverage.

    :param atm_containers: List[elisa.atm.AtmDataContainer];
    :param left_bandwidth: float;
    :param right_bandwidth: float;
    :param kwargs:
    :**kwargs options**:
        * **global_left** * -- float; global wavelength from left where flux for all supllied atmposhperes exist
        * **global_right** * -- float; global wavelength from right where flux for all supllied atmposhperes exist
    :return: List[elisa.atm.AtmDataContainer]
    """
    return [strip_atm_container_by_bandwidth(atm_container, left_bandwidth, right_bandwidth, **kwargs)
            for atm_container in atm_containers]


def strip_atm_container_by_bandwidth(atm_container, left_bandwidth, right_bandwidth, **kwargs):
    """
    Strip atmosphere container model by given bandwidth.
    Usually is model in container defined somewhere in between of left and right bandwidth, never exactly in such
    wavelength. To strip container exactly on bandwidth wavelength, interpolation has to be done. In case, when
    model of any atmosphere has smaller real bandwidth, than bandwidth defined by arguments `right_bandwidth` and
    `left_bandwidth` (it happens in case of bolometric passband), global bandwidth of given atmospheres is used.
    Right gloal bandwidth is obtained as min of all maximal wavelengts from all models and left is max of all mins.


    :param atm_container: elisa.atm.AtmDataContainer;
    :param left_bandwidth: float;
    :param right_bandwidth: float;
    :param kwargs:
    :**kwargs options**:
        * **global_left** * -- float; global wavelength from left where flux for all supllied atmposhperes exist
        * **global_right** * -- float; global wavelength from right where flux for all supllied atmposhperes exist
        * **inplace** * -- bool; if set to True; instead of creation of new DataFrames in
                                 elisa.atm.AtmDataContainers, just existing is inplaced (changed)

    :return: elisa.atm.AtmDataContainer;
    """
    inplace = kwargs.get('inplace', False)
    if atm_container.is_empty():
        ValueError('Atmosphere container is empty.')

    # evaluate whether use argument bandwidth or global bandwidth
    # use case when use global bandwidth is in case of bolometric `filter`, where bandwidth in observer
    # is set as generic left = 0 and right sys.float.max
    atm_model = atm_container.model

    if atm_model.wavelength.min() > left_bandwidth or atm_model.wavelength.max() < right_bandwidth:
        _min, _max = find_global_atm_bandwidth([atm_container])
        # use `global_left` if defined (min of wavelengts where exists intersection of atmospheric models)
        #   or current model left wavelength boundary
        # use `global_righ` if defined (max of wavelengts where exists intersection of atmospheric models) or current
        #   or current model right wavelength boundary
        left_bandwidth, right_bandwidth = kwargs.get("global_left", _min), kwargs.get("global_right", _max)

        if not kwargs.get('global_left') or not kwargs.get('global_right'):
            warnings.warn(f"argument bandwidth is out of bound for supplied atmospheric model\n"
                          f"to avoid interpolation error in boundary wavelength, bandwidth was defined as "
                          f"max {_max} and min {_min} of wavelengt in given model table\n"
                          f"it might leads to error in atmosphere interpolation\n"
                          f"to avoid this problem, please specify global_left and global_right bandwidth as "
                          f"kwargs for given method and make sure all models wavelengths "
                          f"are greater or equal to such limits")
    return strip_to_bandwidth(atm_container, left_bandwidth, right_bandwidth, inplace=inplace)


def strip_to_bandwidth(atm_container, left_bandwidth, right_bandwidth, inplace=False):
    """
    Function directly strips atm container to given bandwidth.

    :param atm_container: elisa.atm.AtmDataContainer; atm container to strip
    :param left_bandwidth: float;
    :param right_bandwidth: float;
    :param inplace: if True `atm_container` is overwritten by striped atmosphere container
    :return: elisa.atm.AtmDataContainer;
    """
    # indices in bandwidth
    valid_indices = list(np.where(np.logical_and(np.greater(atm_container.model.wavelength, left_bandwidth),
                                                 np.less(atm_container.model.wavelength, right_bandwidth)))[0])

    # extend left and right index (left - 1 and right + 1)
    left_extention_index = valid_indices[0] - 1 if valid_indices[0] >= 1 else 0
    right_extention_index = valid_indices[-1] + 1 \
        if valid_indices[-1] < atm_container.model.last_valid_index() else valid_indices[-1]
    atm_cont = atm_container if inplace else deepcopy(atm_container)
    atm_cont.model = atm_cont.model[np.unique([left_extention_index] + valid_indices + [right_extention_index])]
    return extend_atm_container_on_bandwidth_boundary(atm_cont, left_bandwidth, right_bandwidth)


def find_global_atm_bandwidth(atm_containers):
    """
    Function finds common wavelength coverage of the atmosphere models.
    Find intersection of wavelengths of models.

    :param atm_containers: elisa.atm.AtmDataContainer;
    :return: Tuple[float, float]; minimum, maximum wavelength of common coverage (in Angstrom)
    """
    bounds = np.array([[atm.model.wavelength.min(), atm.model.wavelength.max()] for atm in atm_containers])
    return bounds[:, 0].max(), bounds[:, 1].min()


def extend_atm_container_on_bandwidth_boundary(atm_container, left_bandwidth, right_bandwidth):
    """
    Function crops the wavelength boundaries of the atmosphere model to the precise boundaries defined by
    `left_bandwidth` and `right_bandwidth`.

    :param atm_container: elisa.atm.AtmDataContainer;
    :param left_bandwidth: float;
    :param right_bandwidth: float;
    :return: elisa.atm.AtmDataContainer;
    """
    interpolator = interpolate.Akima1DInterpolator(atm_container.model.wavelength, atm_container.model.flux)

    # interpolating values precisely on the border of the filter(s) coverage
    on_border_flux = interpolator([left_bandwidth, right_bandwidth])
    if np.isnan(on_border_flux).any():
        raise AtmosphereError('Interpolation on bandwidth boundaries leed to NaN value.')
    atm_model: AtmModel = atm_container.model
    atm_model.wavelength[np.array([0, -1])] = [left_bandwidth, right_bandwidth]
    atm_model.flux[np.array([0, -1])] = [on_border_flux[0], on_border_flux[1]]
    atm_model.flux = np.round(atm_model.flux, 10)

    atm_container.model = atm_model
    # continute here
    return atm_container


def apply_passband(atm_containers, passband, **kwargs):
    """
    Function applies passband response functions to the stripped atmosphere models.

    :param atm_containers: elisa.atm.AtmDataContainer;
    :param passband: Dict[str, PassbandContainer];
    :return: Dict[str, elisa.atm.AtmDataContainer];
    """
    passbanded_atm_containers = dict()
    logger.debug("applying passband functions on given atmospheres")

    for band, band_container in passband.items():
        if band in ['bolometric']:
            band_container.left_bandwidth = kwargs.get('global_left', band_container.left_bandwidth)
            band_container.right_bandwidth = kwargs.get('global_right', band_container.right_bandwidth)

        passbanded_atm_containers[band] = list()
        for atm_container in atm_containers:
            # strip atm container on passband bandwidth (reason to do it is, that container
            # is stripped on maximal bandwidth defined by all bands, not just by given single band)
            atm_container = strip_to_bandwidth(
                atm_container=deepcopy(atm_container),
                left_bandwidth=band_container.left_bandwidth,
                right_bandwidth=band_container.right_bandwidth,
                inplace=False
            )

            # found passband throughput on atm defined wavelength
            passband_throughput = np.nan_to_num(band_container.akima(atm_container.model.wavelength))

            atm_container.model.flux *= passband_throughput
            passbanded_atm_containers[band].append(atm_container)
    logger.debug("passband application finished")
    return passbanded_atm_containers


def build_atm_validation_hypertable(atlas):
    """
    Prepare validation hypertable to validate atmospheric model (whether is in interpolation bounds).

    :param atlas: str;
    :return: Dict;
    """
    atlas = validated_atlas(atlas)
    all_files = get_list_of_all_atm_tables(atlas)
    filenames = (os.path.basename(f) for f in all_files)
    quantities = sorted([parse_domain_quantities_from_atm_table_filename(f) for f in filenames], key=lambda x: x[0])
    temp_qroups = itertools.groupby(quantities, key=lambda x: x[0])
    hypertable = {
        str(int(temp_qroup[0])):
            {
                "gravity": sorted(set(np.array(list(temp_qroup[1])).T[1])),
                # mettalicity is stored in this dict just because of consitency
                "metallicity": atm_file_prefix_to_quantity_list("metallicity", atlas)
            }
        for temp_qroup in temp_qroups
    }
    return hypertable


def is_out_of_bound(in_arr, values, tolerance):
    """
    Figure out whether `values` are in `in_arr`. Use `tolerance` if you there is allowed.

    :param in_arr: numpy.array;
    :param values: numpy.array;
    :param tolerance: float;
    :return: List[bool];
    """
    values = [values] if not isinstance(values, Iterable) else values
    top, bottom = max(in_arr) + tolerance, min(in_arr) - tolerance
    return [False if bottom <= val <= top else True for val in values]


# pay attention to those methods bellow
# in the future for different atm model might happen that function won't be valid anymore
def validate_temperature(temperature, atlas, _raise=True):
    """
    Validate `temperature`s for existing `atlas`.

    :param temperature: numpy.array;
    :param atlas: str;
    :param _raise: bool; if True, raise ValueError
    :return: bool;
    """
    atlas = validated_atlas(atlas)
    allowed = sorted(atm_file_prefix_to_quantity_list("temperature", atlas))
    invalid = any([True if (allowed[-1] < t or t < allowed[0]) else False for t in temperature])
    if invalid:
        if _raise:
            msg = "Any temperature in system atmosphere is out of bound; it is usually caused by invalid physical model"
            raise TemperatureError(msg)
        return False
    return True


def validate_metallicity(metallicity, atlas, _raise=True):
    """
    Validate `metallicity`s for existing `atlas`.

    :param metallicity: float;
    :param atlas: float;
    :param _raise: bool; if True, raise ValueError
    :return: bool;
    """
    out_of_bound_tol = 0.1  # how far `out of bound` can any value of metallicity runs
    atlas = validated_atlas(atlas)
    allowed = sorted(atm_file_prefix_to_quantity_list("metallicity", atlas))
    out_of_bound = is_out_of_bound(allowed, metallicity, out_of_bound_tol)
    if any(out_of_bound):
        if _raise:
            raise MetallicityError(f"Any metallicity in system atmosphere is out of bound, allowed values "
                                   f"are in range <{min(allowed) - out_of_bound_tol}, {max(allowed) + out_of_bound_tol}"
                                   f">; it is usually caused by invalid physical model")
        return False
    return True


def validate_logg_temperature_constraint(temperature, log_g, atlas, _raise=True):
    """
    Validate `logg`s for existing `atlas` and `temperature`.

    :param temperature: numpy.array;
    :param log_g: numpy.array;
    :param atlas: str;
    :param _raise: bool; if True, raise ValueError
    :return: bool;
    """
    # it has a different name because there is a different interface
    validation_hypertable = build_atm_validation_hypertable(atlas)
    allowed = sorted(atm_file_prefix_to_quantity_list("temperature", atlas))

    invalid = [
        is_out_of_bound(validation_hypertable[
                            str(int(utils.find_nearest_value(allowed, t)[0]))
                        ]["gravity"], [g], 0.1)[0] for t, g in zip(temperature, log_g)]
    if np.any(invalid):
        if _raise:
            raise GravityError("Any gravity (log_g) in system atmosphere is out of bound; "
                               "it is usually caused by invalid physical model")
        return False
    return True


def validate_atm(temperature, log_g, metallicity, atlas, _raise=True):
    """
    Validate atmosphere.
    Run methods::

        validate_temperature
        validate_metallicity
        validate_logg_temperature_constraint

    If anything is not right and `_raise` set to True, raise ValueError.

    :param temperature: numpy.array;
    :param log_g: numpy.array;
    :param metallicity: float;
    :param atlas: str;
    :param _raise: bool; if True, raise ValueError
    :return: bool;
    """
    try:
        metallicity = [metallicity] * len(temperature) if not isinstance(metallicity, Iterable) else metallicity
        validate_temperature(temperature, atlas)
        validate_metallicity(metallicity, atlas)
        validate_logg_temperature_constraint(temperature, log_g, atlas)
    except (ElisaError, ValueError):
        if not _raise:
            return False
        raise
    return True


def atm_file_prefix_to_quantity_list(qname: str, atlas: str):
    """
    Get list of available values for given atm domain quantity, e.g. list of temperatures available in atlas CK04.

    :param qname: str; e.g. `temperature`, `metallicity`, `gravity`
    :param atlas: str; e.g. `castelli` or `ck04`
    :return: List
    """
    atlas = validated_atlas(atlas)
    return getattr(const, f"{str(atlas).upper()}_{str(settings.ATM_DOMAIN_QUANTITY_TO_VARIABLE_SUFFIX[qname])}")


def validated_atlas(atlas):
    """
    Get validated atm atlas, e.g. `castelli` or `ck04` transform to `ck`, it matches folder
    and file prefix for given atlas.

    :param atlas: str; e.g. `castelli` or `ck04`
    :return: str;
    """
    try:
        return settings.ATM_ATLAS_NORMALIZER[atlas]
    except KeyError:
        raise KeyError(f'Incorrect atlas. Following are allowed: {", ".join(settings.ATM_ATLAS_NORMALIZER.keys())}')


def parse_domain_quantities_from_atm_table_filename(filename: str):
    """
    Parse filename to given quantities, e.g. ckm05_3500_g15.csv parse to tuple (-0.5, 3500, 1.5)

    :param filename: str;
    :return: Tuple[float, float, float];
    """
    return get_temperature_from_atm_table_filename(filename), get_logg_from_atm_table_filename(
        filename), get_metallicity_from_atm_table_filename(filename)


def get_metallicity_from_atm_table_filename(filename):
    """
    Get metallicity as number from filename / directory.

    :param filename: str;
    :return: float;
    """
    m = str(filename).split("_")[0][-3:]
    sign = 1 if str(m).startswith("p") else -1
    value = float(m[1:]) / 10.0
    return value * sign


def get_temperature_from_atm_table_filename(filename):
    """
    Get temperature from filename / directory name.

    :param filename: str;
    :return: float;
    """
    return float(str(filename).split("_")[1])


def get_logg_from_atm_table_filename(filename):
    """
    Get logg from filename / directory name.

    :param filename: str;
    :return: float;
    """
    filename = filename if not str(filename).endswith(".csv") else str(filename).replace('.csv', '')
    g = str(filename).split("_")[2][1:]
    return int(g) / 10.0


def get_atm_table_filename(temperature, log_g, metallicity, atlas):
    """
    Get filename based on given descriptive values.

    :param temperature: float;
    :param log_g: float;
    :param metallicity: float;
    :param atlas: str; e.g. `castelli` or `ck04`
    :return: str;
    """
    prefix = validated_atlas(atlas)
    retval = f"{prefix}{utils.numeric_metallicity_to_string(metallicity)}_" \
             f"{int(temperature)}_{utils.numeric_logg_to_string(log_g)}.csv"
    return retval


def get_atm_directory(metallicity, atlas):
    """
    Get table directory name based on given descriptive values.

    :param metallicity: float
    :param atlas: str; e.g. `castelli` or `ck04`
    :return: str
    """
    prefix = validated_atlas(atlas)
    return f"{prefix}{utils.numeric_metallicity_to_string(metallicity)}"


def get_atm_table(temperature, log_g, metallicity, atlas):
    """
    Get dataframe for flux and wavelengths for given values and atlas.
    (Read csv file)

    :param temperature: float;
    :param log_g: float;
    :param metallicity: float;
    :param atlas: str - e.g. `castelli` or `ck04`
    :return: pandas.DataFrame;
    """
    source = settings.ATLAS_TO_BASE_DIR[atlas]
    directory = get_atm_directory(metallicity, atlas)
    filename = get_atm_table_filename(temperature, log_g, metallicity, atlas)
    path = os.path.join(source, directory, filename) if directory is not None else os.path.join(source, filename)

    if not os.path.isfile(path):
        raise FileNotFoundError(f"there is no file like {path}")
    return pd.read_csv(path, dtype=settings.ATM_MODEL_DATAFARME_DTYPES)


def get_list_of_all_atm_tables(atlas):
    """
    Get list of all available atm table files stored in configured location.

    :param atlas: str; e.g. `castelli` or `ck04`
    :return: List[str];
    """
    source = settings.ATLAS_TO_BASE_DIR[validated_atlas(atlas)]
    matches = list()
    for root, dirnames, filenames in os.walk(source):
        for filename in filenames:
            if filename.endswith(('.csv',)):
                matches.append(os.path.join(root, filename))
    return matches


def multithread_atm_tables_reader(path_queue, error_queue, result_queue):
    """
    Multithread reader of atmosphere csv files.

    :param path_queue: Queue;
    :param error_queue: Queue;
    :param result_queue: Queue;
    """
    while True:
        args = path_queue.get(timeout=1)
        if args == "TERMINATOR":
            break
        if not error_queue.empty():
            break
        index, file_path = args
        if file_path is None:
            # consider put here an empty container
            result_queue.put((index, None))
            continue
        try:
            types = {'flux': FLOAT, 'wave': FLOAT}
            t, l, m = parse_domain_quantities_from_atm_table_filename(os.path.basename(file_path))
            atm_container = AtmDataContainer(pd.read_csv(file_path, dtype=types), t, l, m, file_path)
            result_queue.put((index, atm_container))
        except Exception as we:
            error_queue.put(we)
            break


def multithread_atm_tables_reader_runner(fpaths):
    """
    Run multithread reader of csv files containing atmospheric models.

    :param fpaths: Iterable[str];
    :return: Queue;
    """
    n_threads = settings.NUMBER_OF_THREADS

    path_queue = Queue(maxsize=len(fpaths) + n_threads)
    result_queue = Queue()
    error_queue = Queue()

    threads = list()
    try:
        for index, fpath in enumerate(fpaths):
            if not os.path.isfile(fpath):
                logger.debug(f"accessing atmosphere file {fpath}")
                raise AtmosphereError(f"file {fpath} doesn't exist. Your atmosphere tables are either not properly "
                                      f"installed or atmosphere parameters of your stellar model are not covered by "
                                      f"the currently used table.")
            path_queue.put((index, fpath))

        for _ in range(n_threads):
            path_queue.put("TERMINATOR")

        logger.debug("initialising multithread atm table reader")
        for _ in range(n_threads):
            t = Thread(target=multithread_atm_tables_reader, args=(path_queue, error_queue, result_queue))
            threads.append(t)
            t.daemon = True
            t.start()

        for t in threads:
            t.join()
        logger.debug("atm multithread reader finished all jobs")
    except KeyboardInterrupt:
        raise
    finally:
        if not error_queue.empty():
            raise error_queue.get()
    return result_queue


def compute_normal_radiances(matrices_dict, flux_mult=1.0, wave_mult=1.0):
    """
    Run `compute_normal_intensity` method for each band in `matrices_dict`.

    :param matrices_dict: Dict;
    :param flux_mult: float;
    :param wave_mult: float;
    :return: Dict[str, float];
    """
    return {
        band: compute_normal_intensity(
            spectral_flux=dflux[settings.ATM_MODEL_DATAFRAME_FLUX],
            wavelength=dflux[settings.ATM_MODEL_DATAFRAME_WAVE],
            flux_mult=flux_mult,
            wave_mult=wave_mult
        )
        for band, dflux in matrices_dict.items()
    }


def compute_normal_intensity(spectral_flux, wavelength, flux_mult=1.0, wave_mult=1.0):
    """
    Calculates normal flux for all surface faces.

    :param spectral_flux: numpy.array; interpolated atmosphere models for each face (N_face x wavelength)
    :param wavelength: numpy.array or Series; wavelengths of atmosphere models
    :param flux_mult: float;
    :param wave_mult: float;
    :return: numpy.array;
    """
    return flux_mult * wave_mult * integrate.simps(spectral_flux, wavelength, axis=1)


def compute_integral_si_intensity_from_passbanded_dict(passbaned_dict):
    return {
        band: compute_integral_si_intensity_from_atm_data_containers(passbanded_atm)
        for band, passbanded_atm in passbaned_dict.items()
    }


def compute_integral_si_intensity_from_atm_data_containers(atm_data_containers):
    """
    Returns intensity from given atmosphere models.
    If models are already strip by passband, result is also striped

    :param atm_data_containers: Iterable[elisa.atm.AtmDataContainer];
    :return: List[elisa.atm.AtmDataContainer]; integrated `flux` from each elisa.atm.AtmDataContainer
                                               on `wave` in given container
    """
    return [
        IntensityContainer(
            intensity=const.PI * integrate.simps(adc.model.flux * adc.flux_to_si_mult,
                                                 adc.model.wavelength * adc.wave_to_si_mult),
            temperature=adc.temperature,
            log_g=adc.log_g,
            metallicity=adc.metallicity
        )
        for adc in atm_data_containers
    ]


def unique_atm_fpaths(fpaths):
    """
    Group atm table names and return such set and map to origin list.

    :param fpaths: List[str];
    :return: Tuple[str, Dict];

    ::

        (path set - set of unique atmosphere file names,
         map - Dict where every unique atm file has listed indices where it occures)
    """
    fpaths_set = set(fpaths)
    fpaths_map = {key: list() for key in fpaths_set}
    for idx, key in enumerate(fpaths):
        fpaths_map[key].append(idx)
    return fpaths_set, fpaths_map


def remap_passbanded_unique_atms_to_origin(passbanded_containers, fpaths_map):
    """
    Remap atm containers in supplied order by `fpaths_map`.

    :param passbanded_containers: Dict[str, elisa.atm.AtmDataContainer]
    :param fpaths_map: Dict[str, List[int]]; map
    :return: Dict[str, List];
    """
    return {band: remap_unique_atm_container_to_origin(atm, fpaths_map) for band, atm in passbanded_containers.items()}


def remap_unique_atm_container_to_origin(models, fpaths_map):
    """
    Remap atm container in supplied order by `fpaths_map`.

    :warning: assigned containers are mutable, if you will change content of any container, changes will affect
              any other container with same reference

    :param models: List[AtmDatContainer];
    :param fpaths_map: :param fpaths_map: Dict[str, List[int]]; map
    :return: List[elisa.atm.AtmDataContainer];
    """
    models_arr = np.empty(max(list(itertools.chain.from_iterable(fpaths_map.values()))) + 1, dtype='O')
    for model in models:
        models_arr[fpaths_map[model.fpath]] = model
    return models_arr


def read_unique_atm_tables(fpaths):
    """
    Returns atmospheric spectra from table files which encompass the range of surface parameters on the component's
    surface

    :parma fpaths; List[str];
    :return: Tuple[elisa.atm.AtmDataContainers, Dict[str, List]];

    ::

        (List of unique elisa.atm.AtmDataContainers, map - dict where every unique atm file has listed
        indices where it occures)
    """
    fpaths, fpaths_map = unique_atm_fpaths(fpaths)

    # check if the atm table is in the buffer
    models, load_fpaths = [], []
    for fpath in fpaths:
        if fpath in buffer.ATMOSPHERE_TABLES:
            models.append(buffer.ATMOSPHERE_TABLES[fpath])
        else:
            load_fpaths.append(fpath)

    if len(load_fpaths) > 0:
        result_queue = multithread_atm_tables_reader_runner(load_fpaths)
        loaded_models = [qval[1] for qval in utils.IterableQueue(result_queue) if qval[1] is not None]
        # add loaded atmospheres to atm buffer
        for ii, fpath in enumerate(load_fpaths):
            buffer.ATMOSPHERE_TABLES[fpath] = loaded_models[ii]
        models += loaded_models
    # clean buffer
    buffer.reduce_buffer(buffer.ATMOSPHERE_TABLES)
    return models, fpaths_map


def find_atm_si_multiplicators(atm_containers):
    """
    Get atm flux and wavelength multiplicator from `atm_containers`.
    It assume, all containers have the same multiplicators, so it returns values from first one.

    :param atm_containers: List[AtmDatacontainer];
    :return: Tuple[float, float];

    ::

        (flux multiplicator, wavelength multiplicator)

    """
    for atm_container in atm_containers:
        return atm_container.flux_to_si_mult, atm_container.wave_to_si_mult
    raise ValueError('No valid atmospheric container has been supplied to method.')


def find_atm_defined_wavelength(atm_containers):
    """
    Get wavelength from first container from `atm_containers` list.
    It assume all containers has already aligned wavelengths to same.

    :param atm_containers: Iterable[elisa.atm.AtmDataContainer];
    :return: numpy.array[float];
    """
    for atm_container in atm_containers:
        return atm_container.model.wavelength
    raise AtmosphereError('No valid atmospheric container has been supplied to method.')


def remap_passbanded_unique_atms_to_matrix(passbanded_containers, fpaths_map):
    """
    Run `remap_passbanded_unique_atm_to_matrix` for reach container in `passbanded_containers`.

    :param passbanded_containers: List[];
    :param fpaths_map: Dict[str, List[int]]; map - atmosphere container to faces
    :return: Dict[str, numpy.array];
    """
    return {band: remap_passbanded_unique_atm_to_matrix(atm, fpaths_map) for band, atm in passbanded_containers.items()}


def remap_passbanded_unique_atm_to_matrix(atm_containers, fpaths_map):
    """
    Creating matrix of atmosphere models for each face.

    :param atm_containers: List[elisa.atm.AtmDataContainer]; list of unique atmosphere containers from tables
    :param fpaths_map: Dict[str, List[int]]; map - atmosphere container to faces
    :return: numpy.array; matrix of atmosphere models
    """
    total = max(list(itertools.chain.from_iterable(fpaths_map.values()))) + 1
    wavelengths_defined = find_atm_defined_wavelength(atm_containers)
    wavelengths_length = len(wavelengths_defined)
    models_matrix = up.zeros((total, wavelengths_length))

    for atm_container in atm_containers:
        models_matrix[fpaths_map[atm_container.fpath]] = atm_container.model.flux
    return models_matrix


def correct_normal_radiance_to_optical_depth(normal_radiances, ld_cfs):
    """
    Correcting normal radiance values by increment that will correct inacuracy caused by using too shallow optical depth
    for the middle of the disk. Correction was derived analytically from spherical model.

    :param normal_radiances: Dict; dict(component: dict(filter: normal radiances for each face))
    :param ld_cfs: Dict; dict(component: dict(filter: limb darkening coefficients for each face))
    :return: Dict;
    """
    for star, component_normal_radiances in normal_radiances.items():
        ld_coefficients = ld_cfs[star]['bolometric'].T

        coeff = ld.calculate_integrated_limb_darkening_factor(limb_darkening_law=settings.LIMB_DARKENING_LAW,
                                                              coefficients=ld_coefficients)

        normal_radiances[star] = {
            band: normal_radiance / coeff for band, normal_radiance in component_normal_radiances.items()
        }

    return normal_radiances


def planck_function(wavelegth, temperature):
    """
    Standard Planck funcntion.
    :param wavelegth: Union[float, numpy.array]; wavelengths
    :param temperature: float; temperature
    :return: Union[float, numpy.array]
    """
    h = (2.0 * const.PLANCK_CONST * const.C ** 2) / np.power(wavelegth, 5)
    e = (const.PLANCK_CONST * const.C) / (wavelegth * const.BOLTZMAN_CONST * temperature)
    return h / (np.exp(e) - 1.0)


def get_standard_wavelengths():
    """
    Obtain standard wavelengths used in Castelli-Kurucz tables.
    :return: numpy.array
    """
    with open(os.path.join(settings.DATA_PATH, "wavelength.json"), "r") as f:
        return np.array(json.loads(f.read()))



import os
import numpy as np
import pandas as pd

from scipy import interpolate
from . base.error import LimbDarkeningError
from . logger import getLogger
from . import settings
from . import (
    utils,
    const,
    umpy as up
)
from . buffer import buffer

logger = getLogger(__name__)


def get_metallicity_from_ld_table_filename(filename):
    """
    Get metallicity as number from filename typicaly used in van hame ld tables.

    :param filename: str;
    :return: float;
    """
    filename = os.path.basename(filename)
    m = str(filename).split(".")[-2]
    return utils.numeric_metallicity_from_string(m)


def get_ld_table_filename(passband, metallicity, law=None):
    """
    Get filename with stored coefficients for given passband, metallicity and limb darkening default_law.

    :param passband: str
    :param metallicity: str
    :param law: str; limb darkening default_law (`linear`, `cosine`, `logarithmic`, `square_root`)
    :return: str
    """
    law = law if not utils.is_empty(law) else settings.LIMB_DARKENING_LAW
    return f"{settings.LD_LAW_TO_FILE_PREFIX[law]}.{passband}.{utils.numeric_metallicity_to_string(metallicity)}.csv"


def get_ld_table(passband, metallicity, law=None):
    """
    Get content of van hamme table (read csv file).

    :param passband: str;
    :param metallicity: str;
    :param law: str; in not specified, default default_law specified in `elisa.conf.config` is used
    :return: pandas.DataFrame;
    """
    law = law if not utils.is_empty(law) else settings.LIMB_DARKENING_LAW
    filename = get_ld_table_filename(passband, metallicity, law=law)
    path = os.path.join(settings.LD_TABLES, filename)
    if not os.path.isfile(path):
        raise FileNotFoundError(f"There is no file like {path}.")
    return pd.read_csv(path)


def get_ld_table_by_name(fname):
    """
    Get content of van hamme table defined by filename (assume it is stored in configured directory).

    :param fname: str;
    :return: pandas.DataFrame;
    """
    logger.debug(f"accessing limb darkening file {fname}")
    path = str(os.path.join(settings.LD_TABLES, fname))
    if not os.path.isfile(path):
        raise FileNotFoundError(f"There is no file like {path}.")
    return pd.read_csv(path)


def get_relevant_ld_tables(passband, metallicity, law=None):
    """
    Get filename of van hamme tables for surrounded metallicities and given passband.

    :param law: str; limb darkening default_law (`linear`, `cosine`, `logarithmic`, `square_root`)
    :param passband: str;
    :param metallicity: str;
    :return: List;
    """
    # todo: make better decision which values should be used
    surrounded = utils.find_surrounded(const.METALLICITY_LIST_LD, metallicity)
    files = [get_ld_table_filename(passband, m, law) for m in surrounded]
    return files


def interpolate_on_ld_grid(temperature, log_g, metallicity, passband, author=None):
    """
    Get limb darkening coefficients based on van hamme tables for given temperatures, log_gs and metallicity.

    :param passband: Dict;
    :param temperature: Iterable[float];
    :param log_g: Iterable[float]; values expected in log_SI units
    :param metallicity: float;
    :param author: str; (not implemented)
    :return: pandas.DataFrame;
    """
    if isinstance(passband, dict):
        passband = passband.keys()

    # convert logg from log(SI) to log(cgs)
    log_g = utils.convert_gravity_acceleration_array(log_g, units='log_cgs')

    results = dict()
    logger.debug('interpolating limb darkening coefficients')
    for band in passband:
        interp_band = 'bolometric' if band == 'rv_band' else band
        relevant_tables = get_relevant_ld_tables(passband=interp_band, metallicity=metallicity,
                                                 law=settings.LIMB_DARKENING_LAW)
        csv_columns = settings.LD_LAW_COLS_ORDER[settings.LIMB_DARKENING_LAW]
        all_columns = csv_columns

        df = pd.DataFrame(columns=all_columns)

        # for table in relevant_tables:
        for table in relevant_tables:
            if table in buffer.LD_CFS_TABLES:
                _df = buffer.LD_CFS_TABLES[table]
            else:
                _df = get_ld_table_by_name(table)[csv_columns]
                buffer.LD_CFS_TABLES[table] = _df

            apppend = getattr(df, '_append') if hasattr(df, '_append') else getattr(df, 'append')
            df = apppend(_df)
        buffer.reduce_buffer(buffer.LD_CFS_TABLES)

        df = df.drop_duplicates()

        xyz_domain = df[settings.LD_DOMAIN_COLS].values
        xyz_values = df[settings.LD_LAW_CFS_COLUMNS[settings.LIMB_DARKENING_LAW]].values

        uvw_domain = np.column_stack((temperature, log_g))
        uvw_values = interpolate.griddata(xyz_domain, xyz_values, uvw_domain, method="linear")

        if np.any(up.isnan(uvw_values)):
            raise LimbDarkeningError("Limb darkening interpolation lead to numpy.nan/None value. \n"
                                     "Some of the surface parameters (t_eff, log_g, metallicity) are \n"
                                     "probably outside of the supported range. Change the parameters \n"
                                     "of the stars or use your custom ld coefficients that can be passed in \n"
                                     "Star.limb_darkening_coefficients.")

        results[band] = uvw_values

    logger.debug('limb darkening coefficients interpolation finished')
    return results


def limb_darkening_factor(normal_vector=None, line_of_sight=None, coefficients=None, limb_darkening_law=None,
                          cos_theta=None):
    """
    calculates limb darkening factor for given surface element given by radius vector and line of sight vector

    :param line_of_sight: numpy.array; vector (or vectors) of line of sight (normalized to 1 !!!)
    :param normal_vector: numpy.array; single or multiple normal vectors (normalized to 1 !!!)
    :param coefficients: numpy.array;

    shape::

        - numpy.array[[c0, c2, c3, c4,..., cn]] for linear default_law
        - numpy.array[[c0, c2, c3, c4,..., cn],
                      [d0, d2, d3, c4,..., dn]] for sqrt and log default_law

    :param limb_darkening_law: str;  `linear` or `cosine`, `logarithmic`, `square_root`
    :param cos_theta: numpy.array; if supplied, function will skip calculation of its own cos theta and will disregard
                                   `normal_vector` and `line_of_sight`
    :return: numpy.array; gravity darkening factors, the same type/shape as cos_theta
    """
    if normal_vector is None and cos_theta is None:
        raise ValueError('Normal vector(s) was not supplied.')
    if line_of_sight is None and cos_theta is None:
        raise ValueError('Line of sight vector(s) was not supplied.')
    if coefficients is None:
        raise LimbDarkeningError('Limb darkening coefficients were not supplied.')
    if limb_darkening_law is None:
        raise LimbDarkeningError('Limb darkening rule was not supplied choose from: '
                                 '`linear` or `cosine`, `logarithmic`, `square_root`.')

    if cos_theta is None:
        cos_theta = np.sum(normal_vector * line_of_sight, axis=-1)
    else:
        if cos_theta.ndim == 1:
            cos_theta = cos_theta[:, np.newaxis]

    cos_theta = cos_theta.copy()
    negative_cos_theta_test = cos_theta <= 0
    if limb_darkening_law in ['linear', 'cosine']:
        cos_theta[negative_cos_theta_test] = 0.0
        retval = 1.0 - coefficients + coefficients * cos_theta
        retval[negative_cos_theta_test] = 0.0
    elif limb_darkening_law == 'logarithmic':
        cos_theta_for_log = cos_theta.copy()
        cos_theta[negative_cos_theta_test] = 0.0
        cos_theta_for_log[negative_cos_theta_test] = 1.0
        retval = \
            1.0 - coefficients[:, :1] * (1 - cos_theta) - coefficients[:, 1:] * cos_theta * up.log(cos_theta_for_log)
        retval[negative_cos_theta_test] = 0.0
    elif limb_darkening_law == 'square_root':
        cos_theta[negative_cos_theta_test] = 0.0
        retval = 1.0 - coefficients[:, :1] * (1 - cos_theta) - coefficients[:, 1:] * (1 - up.sqrt(cos_theta))
        retval[negative_cos_theta_test] = 0.0
    else:
        raise LimbDarkeningError("Invalid limb darkening.")
    return retval[:, 0] if retval.shape[1] == 1 else retval


def calculate_integrated_limb_darkening_factor(limb_darkening_law=None, coefficients=None):
    """
    Calculates integrated limb darkening factor D(int) for calculating normal radiance from radiosity obtained by
    interpolation in pre-calculated tables.
    D(int) = integral over hemisphere (D(theta)cos(theta)

    :param limb_darkening_law: str -  `linear` or `cosine`, `logarithmic`, `square_root`
    :param coefficients: numpy.array;
    :return: np.array; - bolometric_limb_darkening_factor (scalar for the whole star)
    """
    if coefficients is None:
        raise LimbDarkeningError('Limb darkening coefficients were not supplied.')
    elif limb_darkening_law is None:
        raise LimbDarkeningError('Limb darkening rule was not supplied choose from: '
                                 '`linear` or `cosine`, `logarithmic`, `square_root`.')

    if limb_darkening_law in ['linear', 'cosine']:
        return const.PI * (1 - coefficients[0, :] / 3)
    elif limb_darkening_law == 'logarithmic':
        return const.PI * (1 - coefficients[0, :] / 3 + 2 * coefficients[1, :] / 9)
    elif limb_darkening_law == 'square_root':
        return const.PI * (1 - coefficients[0, :] / 3 - coefficients[1, :] / 5)


def get_bolometric_ld_coefficients(temperature, log_g, metallicity, custom_ld_coefs=None):
    """
    Obtains necessary LD coefficients for each face.

    :param temperature: numpy.array; triangle-wise
    :param log_g: numpy.array; triangle-wise
    :param metallicity: float;
    :param custom_ld_coefs: Union[ńone, dict]; if None
    :return:
    """
    if custom_ld_coefs is not None:
        if 'bolometric' not in custom_ld_coefs:
            raise ValueError('Please ad `bolometric` limb-darkening coefficients to your '
                             'custom set of limb-darkening coefficients.')
        desired_repeats = (temperature.shape[0], 1)
        coeffs = np.tile(custom_ld_coefs['bolometric'], desired_repeats)
    else:
        coeffs = interpolate_on_ld_grid(temperature, log_g, metallicity, passband=["bolometric"])["bolometric"]

    return coeffs.T



from jsonschema import ValidationError


class YouHaveNoIdeaError(ValidationError):
    pass


class ElisaError(Exception):
    pass


class MaxIterationError(ElisaError):
    pass


class InitialParamsError(ElisaError):
    pass


class AtmosphereError(ElisaError):
    pass


class TemperatureError(ElisaError):
    pass


class LimbDarkeningError(ElisaError):
    pass


class MetallicityError(ElisaError):
    pass


class GravityError(ElisaError):
    pass


class MorphologyError(ElisaError):
    pass


class SpotError(ElisaError):
    pass


class SolutionBubbleException(ElisaError):
    def __init__(self, *args, **kwargs):
        self.solution = kwargs.get('solution')
        super().__init__(*args)



from ... logger import getLogger
from . import c_router, lc_point

logger = getLogger('single_system.curves.lc')


def compute_light_curve_without_pulsations(single, **kwargs):
    """
    Compute light curve for single star objects without pulsations.

    :param single: elisa.single_system.system.SinarySystem;
    :param kwargs: Dict;
    :**kwargs options**:
        * ** passband ** * - Dict[str, elisa.observer.PassbandContainer]
        * ** left_bandwidth ** * - float
        * ** right_bandwidth ** * - float
        * ** position_method ** * - function definition; to evaluate orbital positions

    :return: Dict[str, numpy.array];
    """
    initial_system = c_router.prep_initial_system(single)

    lc_labels = list(kwargs["passband"].keys())
    phases = kwargs.pop("phases")

    args = single, initial_system, phases, lc_point.compute_lc_on_pos, lc_labels
    return c_router.produce_curves_wo_pulsations(*args, **kwargs)


def compute_light_curve_with_pulsations(single, **kwargs):
    """
    Compute light curve for single star objects without pulsations.

    :param single: elisa.single_system.system.SingleSystem;
    :param kwargs: Dict;
    :**kwargs options**:
        * ** passband ** * - Dict[str, elisa.observer.PassbandContainer]
        * ** left_bandwidth ** * - float
        * ** right_bandwidth ** * - float
        * ** position_method ** * - function definition; to evaluate orbital positions

    :return: Dict[str, numpy.array];
    """
    initial_system = c_router.prep_initial_system(single, **dict(build_pulsations=False))

    lc_labels = list(kwargs["passband"].keys())
    phases = kwargs.pop("phases")

    args = single, initial_system, phases, lc_point.compute_lc_on_pos, lc_labels
    return c_router.produce_curves_with_pulsations(*args, **kwargs)



from typing import Union
from . mcmc import CentralRadialVelocity as MCMCCentralRV
from . least_squares import CentralRadialVelocity as LstSqrCentralRV
from . import io_tools
from . summary import (
    fit_rv_summary_with_error_propagation,
    fit_lc_summary_with_error_propagation,
    simple_rv_fit_summary,
    simple_lc_fit_summary
)
from .. params import parameters
from .. params.result_handler import FitResultHandler
from .. params.parameters import BinaryInitialParameters
from ... binary_system.utils import resolve_json_kind
from ... logger import getLogger


logger = getLogger('analytics.binary_fit.rv_fit')

DASH_N = 126


class RVFit(FitResultHandler):
    """
    Class with common methods used during an RV fit.
    """
    def __init__(self):
        super().__init__()
        self.fit_method_instance: Union[RVFitLeastSquares, RVFitMCMC, None] = None

    def coefficient_of_determination(self, model_parameters, data, discretization, interp_treshold):
        """
        Function returns R^2 for given model parameters and observed data.

        :param model_parameters: Dict; set of model parameters in json format
        :param data: Dict[str, RVData]; observed RVs for each component
        :param discretization: float; discretization factor for the primary component
        :param interp_treshold: int; a number of observation points above which the synthetic curves will be calculated
                                     using `interp_treshold` equally spaced points that will be subsequently
                                     interpolated to the desired times of observation
        :return: float; coefficient of determination (1.0 means a perfect fit to the observations)
        """
        b_parameters = parameters.BinaryInitialParameters(**model_parameters)
        b_parameters.validate_rv_parameters()
        args = model_parameters, data, discretization, interp_treshold
        return self.fit_method_instance.coefficient_of_determination(*args)


class RVFitMCMC(RVFit):
    """
    Class for RV fitting using the MCMC method.
    """
    def __init__(self):
        super().__init__()
        self.fit_method_instance = MCMCCentralRV()

        self.flat_chain = None
        self.flat_chain_path = None
        self.normalization = None
        self.variable_labels = None

    def fit(self, x0: BinaryInitialParameters, data, **kwargs):
        """
        Perform MCMC sampling on RVFitMCMC instance.

        :param x0: BinaryInitialParameters; initial info about the model parameters such as status
                                            (fixed, variable, constrained), bounds (prior distribution) and
                                            initial value
        :param data: Dict[RVData]; observational data (radial velocities for both components)
        :param kwargs: Dict; arguments passed to the fitting method (see AnalyticsTask.fit kwargs for MCMC or
                             mcmc.CentralRadialVelocity.fit for further info)
        :return: Dict; optimized model parameters in JSON format
        """
        x0.validate_rv_parameters()
        self.result = self.fit_method_instance.fit(data=data, x0=x0, **kwargs)
        self.flat_result = self.fit_method_instance.flat_result

        self.flat_chain = self.fit_method_instance.last_sampler.get_chain(flat=True)
        self.flat_chain_path = self.fit_method_instance.flat_chain_path
        self.normalization = self.fit_method_instance.normalization
        self.variable_labels = list(self.fit_method_instance.fitable.keys())

        logger.info('Fitting and processing of results finished successfully.')
        self.fit_summary()

        return self.result

    def load_chain(self, filename, discard=0, percentiles=None):
        """
        Function loads MCMC chain along with auxiliary data from json file created after each MCMC run.

        :param percentiles: List; percentile intervals used to generate confidence intervals, provided in form:
                                  [percentile for lower bound of confidence interval, percentile of the centre,
                                  percentile for the upper bound of confidence interval]
        :param discard: int; Discard the first discard steps in the chain as a part of the thermalization phase
                             (default: 0).
        :param filename: str; str; chain identificator or filename (ending with .json) containing the chain
        :return: Tuple[numpy.ndarray, List, Dict]; flattened mcmc chain, labels of variables in `flat_chain` columns,
                                                  {var_name: (min_boundary, max_boundary), ...} dictionary of
                                                  boundaries defined by user for each variable needed
                                                  to reconstruct real values from normalized `flat_chain` array
        """
        return io_tools.load_chain(self, filename, discard, percentiles)

    def fit_summary(self, filename=None, **kwargs):
        """
        Function produces detailed summary about the current RV fitting task with the complete error propagation for RV
        parameters if `propagate_errors` is True

        :param filename: str; path where to store summary
        :param kwargs: Dict;
        :**kwargs options**:
            * :propagate_errors: bool; errors of fitted parameters will be propagated to the rest of EB
                                       parameters (takes a while to calculate)
            * :percentiles: List; percentiles used to evaluate confidence intervals from posterior
                                  distribution of EB parameters in MCMC chain . Used only when if
                                  `propagate_errors` is True.
        """
        propagate_errors = kwargs.get('propagate_errors', False)
        percentiles = kwargs.get('percentiles', [16, 50, 84])
        dimensionless_radii = kwargs.get('dimensionless_radii', True)

        kind_of = resolve_json_kind(data=self.result, _sin=True)
        if not propagate_errors:
            if kind_of in ["community"]:
                simple_rv_fit_summary(self, filename)
            else:
                simple_lc_fit_summary(self, filename, dimensionless_radii=True)
            return

        if kind_of in ["community"]:
            fit_rv_summary_with_error_propagation(self, filename, percentiles)
        else:
            fit_lc_summary_with_error_propagation(self, filename, percentiles, dimensionless_radii=dimensionless_radii)

    def filter_chain(self, **boundaries):
        """
        Filtering MCMC chain down to given parameter intervals. This function is useful in case of bimodal distribution
        of the MCMC chain.

        :param boundaries: Dict; dictionary of boundaries e.g. {'primary@te_ff': (5000, 6000), other parameters ...}
        :return: numpy.array; filtered flat chain
        """
        return io_tools.filter_chain(self, **boundaries)


class RVFitLeastSquares(RVFit):
    """
    Class for LC fitting using the Least-Squares method.
    """
    def __init__(self):
        super().__init__()
        self.fit_method_instance = LstSqrCentralRV()

    def fit(self, x0: BinaryInitialParameters, data, **kwargs):
        """
        Perform Least-Squares optimization on RVFitLeastSquares instance.

        :param x0: BinaryInitialParameters; initial info about the model parameters such as status
                                            (fixed, variable, constrained), bounds (prior distribution) and
                                            initial value
        :param data: Dict[RVData]; observational data (radial velocities for both components)
        :param kwargs: Dict; arguments passed to the fitting method (see AnalyticsTask.fit kwargs for Least-Squares or
                             least_squares.CentralRadialVelocity.fit for further info)
        :return: Dict; optimized model parameters in JSON format
        """
        x0.validate_rv_parameters()
        self.result = self.fit_method_instance.fit(data=data, x0=x0, **kwargs)
        self.flat_result = self.fit_method_instance.flat_result

        logger.info('Fitting and processing of results finished successfully.')
        self.fit_summary()

        return self.result

    def fit_summary(self, path=None, **kwargs):
        """
        Function produces detailed summary of the current RV fitting task.

        :param path: str; path, place to store summary
        :param kwargs:
        :return:
        """
        simple_rv_fit_summary(self, path)



import numpy as np

from scipy import special
from scipy import optimize

# numpy
degrees = np.degrees
radians = np.radians

cos = np.cos
sin = np.sin
tan = np.tan
arctan = np.arctan
arcsin = np.arcsin
arccos = np.arccos
arctan2 = np.arctan2
pi = np.pi
zeros = np.zeros
log10 = np.log10
divide = np.divide
log = np.log
exp = np.exp

isnan = np.isnan
NaN = np.NaN
power = np.power
matmul = np.matmul
multiply = np.multiply
inner = np.inner
arange = np.arange
abs = np.abs
ceil = np.ceil
where = np.where
invert = np.invert
concatenate = np.concatenate
sqrt = np.sqrt
sign = np.sign
sum = np.sum

logical_xor = np.logical_xor
logical_and = np.logical_and
logical_or = np.logical_or
mod = np.mod
equal = np.equal
less = np.less
ones = np.ones
round = np.round

# scipy
sph_harm = special.sph_harm
lpmv = special.lpmv
optimize = optimize



import numpy as np

from jsonschema import (
    validate,
    ValidationError
)
from copy import copy

from .. pypex.poly2d.polygon import Polygon
from .. import units, const
from .. units import DefaultBinarySystemUnits
from .. import settings
from .. import umpy as up
from .. base.error import YouHaveNoIdeaError
from .. binary_system import model
from .. utils import is_empty
from .. base.transform import SystemProperties
from .. binary_system.radius import calculate_side_radius


def potential_from_radius(component, radius, phi, theta, component_distance, mass_ratio, synchronicity):
    """
    Calculate potential given spherical coordinates radius, phi, theta.

    :param component: 'primary` or `secondary`;
    :param radius: float;
    :param phi: float;
    :param theta: float;
    :param component_distance: float;
    :param mass_ratio: float;
    :param synchronicity: float;
    :return: float;
    """
    precalc_fn = model.pre_calculate_for_potential_value_primary if component == 'primary' else \
        model.pre_calculate_for_potential_value_secondary
    potential_fn = model.potential_value_primary if component == 'primary' else \
        model.potential_value_secondary

    precalc_args = (synchronicity, mass_ratio, component_distance, phi, theta)
    args = (mass_ratio, ) + precalc_fn(*precalc_args)
    return potential_fn(radius, *args)


def calculate_phase(time, period, t0, offset=0.5):
    """
    Calculates photometric phase from observations.

    :param time: array;
    :param period: array;
    :param t0: float;
    :param offset: float;
    :return: array;
    """
    return up.mod((time - t0 + offset * period) / period, 1.0) - offset


def faces_to_pypex_poly(t_hulls):
    """
    Convert all faces defined as numpy.array to pypex Polygon class instance.

    :param t_hulls: List[numpy.array];
    :return: List;
    """
    return [Polygon(t_hull, _validity=False) for t_hull in t_hulls]


def pypex_poly_hull_intersection(pypex_faces_gen, pypex_hull: Polygon):
    """
    Resolve intersection of polygons defined in `pypex_faces_gen` with polyogn `pypex_hull`.

    :param pypex_faces_gen: List[pypex.poly2d.polygon.Plygon];
    :param pypex_hull: pypex.poly2d.polygon.Plygon;
    :return: List[pypex.poly2d.polygon.Plygon];
    """
    return [pypex_hull.intersection(poly) for poly in pypex_faces_gen]


def pypex_poly_surface_area(pypex_polys_gen):
    """
    Compute surface areas of pypex.poly2d.polygon.Plygon's.

    :param pypex_polys_gen: List[pypex.poly2d.polygon.Plygon];
    :return: List[float];
    """
    return [poly.surface_area() if poly is not None else 0.0 for poly in pypex_polys_gen]


def hull_to_pypex_poly(hull):
    """
    Convert convex polygon defined by points in List or numpy.array to pypex.poly2d.polygon.Polygon.

    :param hull: Union[List, numpy.array];
    :return: pypex.poly2d.polygon.Plygon;
    """
    return Polygon(hull, _validity=False)


def component_to_list(component):
    """
    Converts component name string into list.

    :param component: str;  If None, `['primary', 'secondary']` will be returned otherwise
                            `primary` and `secondary` will be converted into lists [`primary`] and [`secondary`].
    :return: List[str]
    """
    if component in ["all", "both"]:
        component = ['primary', 'secondary']
    elif component in ['primary', 'secondary']:
        component = [component]
    elif is_empty(component):
        return []
    else:
        raise ValueError('Invalid name of the component. Use `primary`, `secondary`, `all` or `both`')
    return component


def move_sys_onpos(
        init_system,
        orbital_position,
        primary_potential: float = None,
        secondary_potential: float = None,
        on_copy: bool = True,
        recalculate_velocities: bool = False
):
    """
    Prepares a postion container for given orbital position.
    Supplied `system` is not affected (is immutable) if `on_copy` is set to True.

    Following methods are applied::

        system.set_on_position_params()
        system.flat_it()
        system.apply_rotation()
        system.add_secular_velocity()
        system.calculate_face_angles()
        system.apply_darkside_filter()

    :param init_system: elisa.binary_system.container.OrbitalPositionContainer;
    :param orbital_position: collections.namedtuple; elisa.const.Position;
    :param primary_potential: float;
    :param secondary_potential: float;
    :param on_copy: bool;
    :param recalculate_velocities: bool; if True, surface elements velocities are recalculated
                                         (usefull while using apsidal symmetry)
    :return: container; elisa.binary_system.container.OrbitalPositionContainer;
    """
    system = init_system.copy() if on_copy else init_system
    system.set_on_position_params(orbital_position, primary_potential, secondary_potential)
    if recalculate_velocities:
        system.build_velocities(components_distance=orbital_position.distance, component='all')
    system.flat_it()
    system.apply_rotation()
    system.add_secular_velocity()
    system.calculate_face_angles(line_of_sight=const.LINE_OF_SIGHT)
    system.apply_darkside_filter()
    return system


def calculate_rotational_phase(system, component):
    """
    Returns rotational phase with in co-rotating frame of reference.

    :param system: elisa.binary_system.contaier.OrbitalPositionContainer; instance
    :param component: str; `primary` or `secondary`
    :return: float;
    """
    star = getattr(system, component)
    return (star.synchronicity - 1.0) * system.position.phase


def validate_binary_json(data):
    """
    Validate input json to create binary instance from.

    :param data: Dict; json like object
    :return: bool; return True if valid schema, othervise raise error
    :raise: ValidationError;
    """
    schema_std = settings.SCHEMA_REGISTRY.get_schema("binary_system_std")
    schema_community = settings.SCHEMA_REGISTRY.get_schema("binary_system_community")
    std_valid, community_valid = False, False

    try:
        validate(instance=data, schema=schema_std)
        std_valid = True
    except ValidationError:
        pass

    try:
        validate(instance=data, schema=schema_community)
        community_valid = True
    except ValidationError:
        pass

    # previous code cannot catch error when user inputs only one argument from the other parameter input format
    if ('mass_ratio' in data['system'].keys() or 'semi_major_axis' in data['system'].keys()) and std_valid is True:
        raise ValidationError("You probably tried to input your parameters in `standard` format but your "
                              "parameters include `mass ratio` or `semi_major_axis` (use either (M1, M2) or  (q, a)).")

    if ('mass' in data['primary'].keys() or 'mass' in data['secondary'].keys()) and community_valid is True:
        raise ValidationError("You probably tried to input your parameters in `community` format but your "
                              "parameters include masses of the components (useeither (M1, M2) or  (q, a)).")

    if (not community_valid) & (not std_valid):
        raise ValidationError("BinarySystem cannot be created from supplied json schema. ")

    if community_valid & std_valid:
        raise YouHaveNoIdeaError("Make sure that list of fitted parameters contain only `standard` or `community` "
                                 "combination of parameter (either (M1, M2) or  (q, a)).")

    return True


def resolve_json_kind(data, _sin=False):
    """
    Resolve if json is `std` or `community`.

    std - standard physical parameters (M1, M2)
    community - astro community parameters (q, a)

    :param data: Dict; json like
    :param _sin: bool; if False, looking for `semi_major_axis` in given JSON, otherwise looking for `asini`
    :return: str; `std` or `community`
    """
    lookup = "asini" if _sin else "semi_major_axis"
    m1, m2 = data.get("primary", dict()).get("mass"), data.get("secondary", dict()).get("mass")
    q, a = data["system"].get("mass_ratio"), data["system"].get(lookup)

    if m1 and m2:
        return "std"
    if q and a:
        return "community"
    raise LookupError("It seems your JSON is invalid.")


def transform_json_community_to_std(data):
    """
    Transform `community` input json to `std` json.
    Compute `M1` and `M2` from `q` and `a`.

    :param data: Dict;
    :return: Dict;
    """
    q = data["system"].pop("mass_ratio")
    a = SystemProperties.semi_major_axis(data["system"].pop("semi_major_axis"))
    period_val = copy(data["system"]["period"])
    period = (SystemProperties.period(period_val) * DefaultBinarySystemUnits.system.period).to(units.TIME_UNIT).value
    m1 = ((4.0 * const.PI ** 2 * a ** 3) / (const.G * (1.0 + q) * period ** 2))
    m1 = np.float64((m1 * units.kg).to(units.solMass))
    m2 = q * m1

    data["primary"].update({"mass": m1})
    data["secondary"].update({"mass": m2})

    return data


def correction_to_com(distance, mass_ratio, scom):
    """
    Calculates the correction for com from primary-centered coordinate system to barycentric.

    :param distance: float;
    :param mass_ratio: float
    :param scom: float; secondary component component of mass
    :return: float; correction to com in primary-centered system
    """
    distances_to_com = distance * mass_ratio / (1 + mass_ratio)
    dir_to_secondary = scom / np.linalg.norm(scom)
    return distances_to_com * dir_to_secondary


def calculate_sma_estimate(mass_ratio, synchronicity, potential, period, component, mid_g):
    """
    Estimates a value of semi major axis in a manner that puts average surface
    gravity of the components to the value `mid_g`. Very useful in case of LC
    fit where sma has to be fixed to some sensible value.

    :param mass_ratio: float; M2/M1
    :param synchronicity: float;
    :param potential: float; surface potential
    :param period: float; orbital period in days
    :param component: str; `primary` or `secondary`
    :param mid_g: float; desired average surface acceleration in m.s^-2
    :return: float; sma estimate in solar radii
    """
    radius = calculate_side_radius(synchronicity, mass_ratio, 1.0, potential, component)
    q_funcval = 1.0 / (1. + mass_ratio) if component == 'primary' else mass_ratio / (1. + mass_ratio)
    return 1.4374e-9 * mid_g * (radius * 86400 * period)**2 / (4 * const.PI ** 2 * q_funcval)
